# -*- coding: utf-8 -*-
# Copyright (c) 2000-2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Validate projman XML formats
"""

# ------------------------------------------------------------
# The following section should be generic enough to go into
# its own library something like logilab.common.xmlcheck
# ------------------------------------------------------------

import mx.DateTime
try:
    import xml.etree.ElementTree as ET
except ImportError:
    import elementtree.ElementTree as ET


def any(attrib, attr):
    return ""

def not_empty(attrib, attr):
    if not attrib[attr]:
        return "%s must not be the empty string" % attr
    return ""

def iso_date(dt):
    if len(dt) == 10:
        return mx.DateTime.strptime( dt, "%F" )
    elif len(dt) == 22:
        return mx.DateTime.strptime( dt, "%F %H:%M:%S.%S" )
    else:
        print "date = ", dt, "lg=", len(dt)
        raise RuntimeError("date non valide")

def iso_time(tm):
    date = mx.DateTime.strptime( tm, "%H%M" )
    return mx.DateTime.Time(date.hour, date.minute)

class attribute_checker(object):
    def __call__(self, attrib, attr):
        if attr not in attrib:
            assert RuntimeError("Framework should make sure the attribute is present"
                                "before calling the checker" )
        val = attrib[attr]
        return self.check(attrib, attr, val)

class convertible(attribute_checker):
    def __init__(self, cvt, msg=None):
        self.cvt = cvt
        if msg is None:
            msg = "Cannot convert %%(attr)s:%%(val)r using %s" % cvt
        self.msg = msg

    def check(self, attrib, attr, val):
        try:
            self.cvt(val)
        except:
            
            return self.msg % {'attr':attr, 'val':val}

class depends(object):
    def __init__(self, *attrs):
        self.attrs = attrs

    def __call__(self, attrib, attr):
        if attr not in attrib:
            assert RuntimeError("Framework should make sure the attribute is present"
                                "before calling the checker" )
        # check for the other attributes to be present with this one
        for a in self.attrs:
            if a not in attrib:
                return "%s must be specified along with : %s" % (attr, ", ".join(self.attrs))
        return None

class matches(attribute_checker):
    def __init__(self, txt):
        import re
        self.txt = txt
        self.re = re.compile(txt)

    def check(self, attrib, attr, val):
        if not self.re.search(val):
            return "attribute %s does not match %r" % (attr, self.txt)

class one_of(attribute_checker):
    def __init__(self, *args):
        self.values = args

    def check(self, attrib, attr, val):
        if val not in self.values:
            return "attribute %s must be one of %s (not %s)" % (attr, ",".join(self.values), repr(val))

class BaseEtreeChecker(object):
    def __init__(self):
        self._errors = []
        self._stack = []
        self._ignore_children_flag = False

    def _ignore_children(self):
        self._ignore_children_flag = True

    def check(self):
        """Generic validation methods lookup
        overide this if you have special cases
        like a 'check' node or other special chars
        """
        pos, node = self.stack[-1]
        tag = node.tag.replace("-","_")
        checker = getattr(self, "check_"+tag, None)
        self._ignore_children_flag = False
        if checker:
            checker()
        else:
            self._error("Unknown tag %s" % node.tag)
        return self._ignore_children_flag

    def _error(self, msg):
        pos = [ p for p,n in self.stack ]
        pos = [0]+pos[:-1]
        tag = [ n.tag for p,n in self.stack]
        path = "/".join( [ t+"[%s]"%p for t,p in zip(tag,pos) ] )
        self._errors.append( "%s:%s" % (path,msg) )

    def __str__(self):
        return "\n".join(self._errors)

    def validate(self, tree, filename=''):
        """Validate the source file format"""
        self.stack = [ [-1,tree.getroot()] ]
        # Do a DFS exploration so that we always have
        # the stack match the path of the element
        while self.stack:
            pos, node = self.stack[-1]
            if pos==-1: # first visit
                ignore_children = self.check()
                if ignore_children:
                    pos=len(node)

            if pos+1<len(node):
                self.stack[-1][0]+=1
                new_node = node[pos+1]
                self.stack.append( [-1, new_node] )
                continue
            # we finished all children
            self.stack.pop(-1)
        if self._errors:
            return False
        else:
            return True

    # various helpers
    def _isroot(self):
        self._is_child_of( None )

    def _noattr(self):
        pos, node = self.stack[-1]
        if node.attrib:
            self._error( "%s should not have attributes" % node.tag )

    def _is_child_of(self, *parents):
        pos, node = self.stack[-1]
        if len(self.stack)>=2 and self.stack[-2][1].tag not in parents:
            self._error( "%s must be a child of %s" % (node.tag, parents) )
        if len(self.stack)==1 and None not in parents:
            self._error( "%s must be the root node" % node.tag )


    def _attributes(self, attribs):
        pos, node = self.stack[-1]
        attrs = {}
        for name, checkers in attribs.items():
            opt = False
            if name.endswith("?"):
                name = name[:-1]
                opt = True
            attrs[name]=True
            if name not in node.attrib:
                if not opt:
                    self._error("attribute %s is missing" % name)
                continue
            if not isinstance(checkers, (list, tuple)):
                checkers = (checkers,)
            for c in checkers:
                msg = c(node.attrib, name)
                if msg:
                    self._error( msg )
        for name in node.attrib.keys():
            if name not in attrs:
                self._error("unknown attribute '%s'" % name)

    def _parent_node(self):
        if len(self.stack)<=1:
            return ET.Element("")
        return self.stack[-2][1]

    def _children(self, *types):
        pos, node = self.stack[-1]
        children = {}
        for t in types:
            if t[-1] in '?*+':
                t = t[:-1]
            children[t] = 0
        for c in node:
            if c.tag not in children:
                self._error("%s should not have child %s" % (node.tag, c.tag))
            else:
                children[c.tag]+=1
        for t in types:
            if t[-1] in '?*+':
                t0=t[:-1]
            else:
                t0=t
            n = children[t0]
            if t[-1] == '?' and n>1:
                self._error("%s should appear at most once instead of %d" % (t0, n))
            elif t[-1] == '*':
                pass
            elif t[-1] == '+' and n==0:
                self._error("%s should appear at least once instead of %d" % (t0, n))
            elif t[-1] not in '?*+' and n!=1:
                self._error("%s should appear exactly once instead of %d" % (t0, n))

    def _content(self, cvt, msg=None):
        pos, node = self.stack[-1]
        txt = node.text or ""
        if msg is None:
            msg = "%s should convert to %s" % (node.tag,cvt)
        for c in node:
            txt += c.tail or ""
        try:
            cvt(txt.strip())
        except:
            self._error(msg % {'txt':repr(txt)})
            import traceback
            traceback.print_exc()
            
    def _empty(self):
        pos, node = self.stack[-1]
        def isempty(txt):
            assert txt==""
        self._content(isempty, "should not have text : %(txt)s")

# -------------------------------------------------------------------------
# The following are checkers that check the validity of projman's project
# files
# -------------------------------------------------------------------------

class ProjectChecker(BaseEtreeChecker):
    def check_project(self):
        self._isroot()
        self._noattr()
        self._empty()
        self._children( "import-tasks",
                            "import-resources",
                            "import-schedule",
                            "import-activities", )
    def check_import_(self):
        self._is_child_of("project")
        self._attributes( { "file": not_empty } )
        self._empty()

    check_import_tasks = check_import_
    check_import_activities = check_import_
    check_import_schedule = check_import_
    check_import_resources = check_import_

DATE_CONSTRAINT = ["begin-at-date", "end-at-date",
                   "begin-after-date", "end-after-date",
                   "begin-before-date", "end-before-date",
                   ]
TASK_CONSTRAINT = ["begin-after-end", "end-after-end",
                   "begin-after-begin", "end-after-begin",
                   "synchronized",]

class ScheduleChecker(BaseEtreeChecker):
    def check_schedule(self):
        self._isroot()
        self._noattr()
        self._children( "task+", "milestone*" )

    def check_task(self):
        self._is_child_of("schedule")
        self._attributes( {"id": not_empty} )
        pos, node = self.stack[-1]
        self._children("costs_list", "report-list",
                       "global-cost", "constraint-date*",
                       "status", "constraint-task*")

    def check_constraint_date(self):
        self._is_child_of("task","milestone")
        self._children()
        if self._parent_node().tag == "milestone":
            self._attributes( {"type" : one_of("at-date") } )
        else:
            self._attributes( {"type" : one_of(*DATE_CONSTRAINT)} )
        self._content( iso_date )
        
    def check_constraint_task(self):
        self._empty()
        self._children()
        self._attributes( {"type": one_of(*TASK_CONSTRAINT),
                           "idref": not_empty} )

    def check_status(self):
        self._content( str )
        self._children()
        self._noattr()

    def check_report(self):
        self._children()
        self._empty()
        self._attributes( {"usage": convertible(float),
                           "to":  convertible(iso_date),
                           "idref": any,
                           "from": convertible(iso_date)} )
    def check_cost(self):
        self._children()
        self._content( float )
        self._attributes( {"idref" : not_empty} )
    def check_global_cost(self):
        self._children()
        self._content( float )
        self._attributes( {"unit": any} )
    def check_priority(self):
        self._error("TODO")
    def check_costs_list(self):
        self._empty()
        self._noattr()
        self._children( "cost+" )
    def check_report_list(self):
        self._empty()
        self._noattr()
        self._children( "report+" )
    def check_milestone(self):
        self._is_child_of('schedule')
        self._empty()
        self._attributes({"id": not_empty})
        self._children('constraint-date', 'constraint-task*')

class ResourcesChecker(BaseEtreeChecker):
    def check_resources_list(self):
        self._isroot()
        self._attributes( { "id?" : not_empty } )
        self._children("resource+","calendar+","resource-role*")
        self._empty()
        
    def check_resource(self):
        self._is_child_of( "resources-list" )
        self._children("label","use-calendar","hourly-rate?", "role*")
        self._empty()
        self._attributes( {"type?" : not_empty,
                           "id" : not_empty,} )

    def check_label(self):
        self._children()
        self._noattr()
        self._is_child_of("resource","resource-role","calendar","day-type")
        self._content( unicode )

    def check_role(self):
        self._children()
        self._empty()
        self._is_child_of("resource")
        self._attributes({'idref' : not_empty})

    def check_use_calendar(self):
        self._children()
        self._empty()
        self._is_child_of( "resource" )
        self._attributes( {"idref": not_empty} )

    def check_hourly_rate(self):
        self._children()
        self._content( float )
        self._is_child_of( "resource" )
        self._attributes( {"unit?": one_of("euros")} )

    def check_resource_role(self):
        self._is_child_of( "resources-list" )
        self._children("label")
        self._empty()
        self._attributes( {"id" : not_empty,
                           "hourly-cost" : convertible(float),
                           "cost-unit?" : one_of("EUR")} )

    def check_calendar(self):
        self._is_child_of( "calendar", "resources-list" )
        self._empty()
        self._attributes( {"id": not_empty} )
        self._children("calendar*", "label","day-types?","day*",
                       "timeperiod*", "start-on?", "stop-on?")

    def check_day_types(self):
        self._is_child_of("calendar")
        self._empty()
        self._children( "day-type+" )
        self._attributes( {"default": not_empty})

    def check_day_type(self):
        self._is_child_of("day-types")
        self._children( "label","interval*" )
        self._empty()
        self._attributes( {"id": not_empty} )

    def check_interval(self):
        self._is_child_of("day-type")
        self._empty()
        self._children()
        self._attributes( {"start": convertible(iso_time),
                           "end": convertible(iso_time)} )

    def check_day(self):
        self._is_child_of("calendar")
        self._content(str)
        self._children()
        self._attributes( {"type": not_empty} )

    def check_timeperiod(self):
        self._is_child_of("calendar")
        self._empty()
        self._children()
        self._attributes( {"to": convertible(iso_date),
                           "from": convertible(iso_date),
                           "type": not_empty} )

    def check_start_on(self):
        self._is_child_of("calendar")
        self._content(iso_date)
        self._children()
        self._noattr()

    def check_stop_on(self):
        self._is_child_of("calendar")
        self._content(iso_date)
        self._children()
        self._noattr()
    
class TasksChecker(BaseEtreeChecker):
    def check_task(self):
        self._is_child_of("task",None)
        self._attributes( {"id":not_empty,
                           "load-type?":(one_of("oneof","shared","sameforall","spread"),
                                         depends("load") ),
                           "load?":(convertible(float),
                                    depends("load-type")),
                           "resource-role?" : not_empty,
                           } )
        self._empty()
        pos, node = self.stack[-1]
        if node.findall("task") or node.findall("milestone"):
            # a task with sub-tasks
            self._children("label", "link?", "description?","constraint-date*",
                           "constraint-resource*","constraint-task*",
                           "task*", "milestone*")
        else:
            # a leaf task
            pos, node = self.stack[-1]
            # we don't want duration with the new load-type descriptions
            self._children("label", "link?", "description?","constraint-date*",
                           "constraint-resource*","constraint-task*",
                           "constraint-interruptible?")
            
    def check_milestone(self):
        self._is_child_of( "task" )
        self._attributes( {"id":not_empty} )
        self._empty()
        self._children("label","description?","constraint-date*",
                       "constraint-task*")

    def check_description(self):
        self._is_child_of("task","milestone")
        self._attributes( {"format?" : one_of("rest","docbook")} )
        self._content(unicode)
        self._ignore_children()

    def check_label(self):
        self._is_child_of("task","milestone")
        self._content(unicode)
        self._noattr()
        self._children()

    def check_link(self):
        self._is_child_of('task', 'milestone')
        self._empty()
        self._attributes({'url' : not_empty})
        self._children()

    def check_constraint_date(self):
        self._is_child_of("task","milestone")
        self._content(iso_date)
        self._attributes( {"type" : one_of(*DATE_CONSTRAINT),
                           "priority?": one_of('1','2','3')} )
        self._children()

    def check_constraint_task(self):
        self._is_child_of("task","milestone")
        self._empty()
        self._attributes( {"type" : one_of(*TASK_CONSTRAINT),
                          "idref" : not_empty,
                          "priority?": one_of('1','2','3') } )
        self._children()

    def check_constraint_resource(self):
        self._is_child_of("task","milestone")
        self._empty()
        self._attributes( {"idref" : not_empty,
                           "usage?" : not_empty,
                           "type?" : not_empty} )
        self._children()

    def check_constraint_interruptible(self):
        self._is_child_of("task","milestone")
        self._empty()
        self._attributes( { "type" : one_of('True', 'False'),
                            "priority" : one_of('1', '2', '3')})

class ActivitiesChecker(BaseEtreeChecker):
    def check_activities(self):
        self._isroot()
        self._empty()
        self._attributes( {"id": not_empty} )
        self._children("reports-list*")

    def check_reports_list(self):
        self._is_child_of("activities")
        self._empty()
        self._attributes( {"task-id": not_empty} )
        self._children("report+")

    def check_report(self):
        self._is_child_of("reports-list")
        self._empty()
        self._attributes({"idref":not_empty,
                          "from":convertible(iso_date),
                          "to":convertible(iso_date),
                          "usage":convertible(float)})
        self._children()



# Copyright (c) 2000-2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""%%prog %s [options] %s

Projman's command line tool. 
%s"""

def run(args):
    from logilab.common.clcommands import main_run
    from projman import commands
    main_run(args, __doc__)



# Copyright (c) 2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""provide classes for projman commands."""

import logging

from logilab.common.clcommands import BadCommandUsage, Command, \
     register_commands

from projman.__pkginfo__ import version
from projman.views import document
from projman.readers import ProjectXMLReader
from projman.writers.projman_writer import write_schedule_as_xml, indent
from projman.views import ALL_VIEWS

# verbosity to logging level mapping
LEVELS = {
    0 : logging.ERROR,
    1 : logging.WARN,
    2 : logging.INFO,
    3 : logging.DEBUG,
    }

class ProjmanCommand(Command):
    """base class providing common behaviour for projman commands"""
    def __init__(self, __doc__):
        Command.__init__(self, __doc__=__doc__, version=version)

    options = (
        ('project-file',
         {'short': 'f',
          'type' : 'string', 'metavar': '<project description file>',
          'default': 'project.xml',
          'help': 'specify the project description file to use',
          }
         ),
        ('verbose',
         {'type' : 'int', 'metavar': '<0..3>',
          'default': 0,
          'help': 'display additional information during execution of projman',
          }
         ),
        ('task-root',
         {'short': 'R',
          'type' : 'string', 'metavar': '<virtual task root>',
          'default': None,
          'help': 'identifier of a task to use as root',
          }
         ),
        ('factorized_days',
         {'type' : 'int', 'metavar' : '<1, 2 or 4>',
          'default': 1,
          'help' : 'schedule and display task in day(1), half day(2) or quarter of day(4).' 
                   'Attention: if you decide to use this option, be coherent ' 
                   'for next projman commands ! '
          }
         ),
        )

    def run(self, args):
        """run the command with its specific arguments"""
        loglevel = LEVELS.get(self.config.verbose, logging.WARN)
        logging.basicConfig(level=loglevel)
        reader = ProjectXMLReader(self.config.project_file, self.config.task_root)
        self.project, self.files = reader.read()
        self.project.factor = self.config.factorized_days
        self._run(args)

    def _run(self, args):
        raise NotImplementedError

# Concrete commands ###########################################################
class CheckCommand(ProjmanCommand):
    "check the definition of a projman problem"
    name = 'check'
    max_args = 0
    arguments = ''

    def _run(self, args):
        from projman.checker.problem_checker import Checker
        # validate xml format
        reader = ProjectXMLReader(self.config.project_file, self.config.task_root)
        # validate projman problem
        check_project = Checker(self.project, self.config.verbose)
        check_project.problem_checker()

class ScheduleCommand(ProjmanCommand):
    """schedule a project"""
    name = 'schedule'
    max_args = 0
    arguments = ''

    options = ProjmanCommand.options + (
        ('type',
         {'type' : 'choice', 'metavar': '<schedule type>',
          'choices': ('dumb', 'simple', 'csp'),
          'default': 'csp',
          'help': 'scheduling method',
          }
         ),
        ('time', 
         {'type' : 'int', 'metavar': '<1...>',
          'default': 400000,
          'help': 'stop the programm after time'
          }
         ),
        )

    def _run(self, views):
        from projman.scheduling import schedule
        from projman.checker.problem_checker import Checker
        # validate projman problem
        check = Checker(self.project, self.config.verbose)
        check.problem_checker()
        if check.errors:
            return
        rounder = 0
        schedule(self.project, self.config)
        while (self.project.nb_solution == 0 and rounder < 2):
            print '\nAttention: No valid solution !'
            print "constraints with priority", self.project.priority,
            print 'are dropped ...\n'
            reader = ProjectXMLReader(self.config.project_file,
                                self.config.task_root)
            self.project, self.files = reader.read()
            self.project.factor = self.config.factorized_days
            rounder += 1
            self.project.priority -= rounder
            schedule(self.project, self.config)
            nb_solution = self.project.nb_solution
        write_schedule_as_xml(self.files['schedule'], self.project)


class ViewCommand(ProjmanCommand):
    """generate XML view(s) from a project file (usually using Documentor +
    docbook dialect)
    """
    name = 'view'
    min_args = 1
    arguments = "|".join(ALL_VIEWS.keys())

    options = ProjmanCommand.options + (
        ('output',
         {'short': 'o',
          'type' : 'string', 'metavar': '<output xml file>',
          'default': 'output.xml',
          'help': 'specific output file to use',
          }
         ),
        ('display-dates',
         {'type' : 'yn', 'metavar': '<y or n>',
          'default': True,
          'help': 'display task\'s begin and end date (tasks-list view only)',
          }
         ),
        ('level',
         {'type': 'int', 'metavar': '<1, ... >',
         'default' : 3,
         'help': 'display tasks on  1, 2, or more level in views tables',
          }
         ),
        )

    def _run(self, views):
        tree = document("root")
        if list(views) == ["all",]:
            views = ALL_VIEWS.keys()
        for viewname in views:
            try:
                viewklass = ALL_VIEWS[viewname]
            except KeyError:
                raise BadCommandUsage('unknown view %s' % viewname)
            view = viewklass(self.config)
            view.generate(tree, self.project)
        output = file(self.config.output, 'w')
        indent(tree.getroot())
        tree.write( output, encoding="UTF-8" )
        output.close()


class DiagramCommand(ProjmanCommand):
    """generate diagrams from a project file (resources, gantt, etc.)"""
    name = 'diagram'
    min_args = 1
    arguments = '<diagram name>...'

    options = ProjmanCommand.options + (
        ('output',
         {'short': 'o',
          'type' : 'string', 'metavar': '<output xml file>',
          'default': None,
          'help': 'specific output file to use when a single diagram is generated',
          }
         ),
        ('selected-resource',
         {'type' : 'string', 'metavar': '<resource identifier>',
          'default': None,
          'help': 'specifies the id of the resource to take in account for '
                  'resources diagrams',
          }
         ),
        ('format',
         {'type' : 'choice', 'metavar': '<format>',
          'choices': ('svg', 'png'), # 'html'  ## png ne marche plus
          'default': 'svg',
          'help': 'specifies the output format for diagrams',
          }
         ),
        ('depth',
         {'type' : 'int', 'metavar': '<level>',
          'default': 0,
          'help': 'specifies the depth to visualize for diagrams, default to '
                  '0 which means all the tree',
          }
         ),
        ('timestep',
         {'type' : 'string', 'metavar': '<day, week, month>',
          'default': 'day',
          'help': 'timeline increment for diagram',
          }
         ),
        ('view-begin',
         {'type' : 'date', 'metavar': '<yyyy/mm/dd>',
          'default': None,
          'help': 'begin date for diagram view',
          }
         ),
        ('view-end',
         {'type' : 'date', 'metavar': '<yyyy/mm/dd>',
          'default': None,
          'help': 'end date for diagram view',
          }
         ),
        ('del-ended',
         {'type' : 'yn', 'metavar': '<y or n>',
          'default': False,
          'help': 'do not display in resource diagram tasks wich are completed, '
                  'meaning that time of work on them equals theirs duration.',
          }
         ),
        ('del-empty',
         {'type' : 'yn', 'metavar': '<y or n>',
          'default': False,
          'help': 'do not display in resource diagram tasks wich are not '
                  'worked during given period',
          }
         ),
        )

    def _run(self, diagrams):
        #if self.config.renderer == 'html':
        #    from projman.renderers.HTMLRenderer import ResourcesHTMLRenderer
        #    renderer = ResourcesHTMLRenderer(self.options.get_render_options())
        #else:
        from projman.renderers import ResourcesRenderer, GanttRenderer, \
             GanttResourcesRenderer, HandlerFactory
        handler = HandlerFactory(self.config.format)
        known_diagrams = {
            'gantt': GanttRenderer,
            'resources': ResourcesRenderer,
            'gantt-resources': GanttResourcesRenderer,
            }
        if self.config.output and len(diagrams)>1:
            print "You specified more than one diagram with an output file name"
            print "*** only the last one will be printed ***"
        for diagram in diagrams:
            try:
                renderer = known_diagrams[diagram](self.config, handler)
            except KeyError:
                raise BadCommandUsage('unknown diagram %s' % diagram)
        if self.config.timestep:
            if not self.config.timestep in ['day', 'month', 'week']:
                raise BadCommandUsage('non valid timestep %s' %self.config.timestep) 
            
            output = self.config.output or '%s.%s' % (diagram, self.config.format)
            stream = handler.get_output(output)
            renderer.render(self.project, stream)



class DiagramCommand2(ProjmanCommand):  # a test...
    """generate diagrams from a project file (resources, gantt, etc.)"""
    name = 'diagram2'
    min_args = 1
    arguments = '<diagram name>...'

    options = ProjmanCommand.options

    def _run(self, diagrams):
        #if self.config.renderer == 'html':
        #    from projman.renderers.HTMLRenderer import ResourcesHTMLRenderer
        #    renderer = ResourcesHTMLRenderer(self.options.get_render_options())
        #else:
        from projman.renderers.gantt2 import GanttRenderer2
        gantt = GanttRenderer2(self.config, self.project)
        gantt.render("gantt2.svg")
        gantt.save()


register_commands((CheckCommand, ScheduleCommand, ViewCommand, DiagramCommand))



# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
rendering resources diagrams
"""

__revision__ = "$Id: resource.py,v 1.4 2006-04-13 13:07:27 nico Exp $"

import logging
log = logging.getLogger()

from logilab.common.tree import NodeNotFound

from projman.lib import date_range
from projman.lib.constants import HOURS_PER_DAY
from projman.renderers.abstract import \
     AbstractRenderer, AbstractDrawer, TODAY, \
     TITLE_COLUMN_WIDTH, FIELD_COLUMN_WIDTH, ROW_HEIGHT
from mx.DateTime import oneHour

class ResourcesRenderer(AbstractRenderer) :

    AbstractRenderer.DEFAULT_OPTIONS.update({'selected-resource' : None})
    
    def __init__(self, options, handler,
                 colors_file=None, colors_stream=None):
        """
        Creates concrete renderer with a matching associate drawer:
        ResourceDrawer
        """
        AbstractRenderer.__init__(self, options)
        self.drawer = ResourcesDrawer(options,
                                      handler, colors_file, colors_stream)
        self._draw_resources = {}
       
    def render(self, project, stream):
        """
        render the task as a resources diagram
        """
        AbstractRenderer.render(self, project, stream)
        self.drawer._handler.save_result(stream)
     
    def _render_body(self, project) :
        """
        Generate events to draw a diagram for the resources activities description.
        
        concrete renderer should use the 'write' method or override this method
        to return the generated content.
        """
        self.drawer.main_title("Resources activities")
        self.drawer.draw_timeline()
        self.drawer.close_line()
        # print the resources informations
        resources = [resource for resource in project.resource_set.children 
                     if resource.TYPE == 'resource']
        if self.options.selected_resource:
            resources = [resource for resource in resources
                         if resource.id == self.options.selected_resource]
        for resource in resources:
            self.render_resource(resource, project)
        
    def render_resource(self, resource, project) :
        """
        generate events for a given resource
        """
        self.drawer.set_color_set(self._i)
        self._i += 1
        self.render_total_occupation(project, resource)
        # print activities for this resource
        grouped = project.activities.groupby('resource','task')
        if resource.id in grouped:
            for t_id, activities in grouped[resource.id].iteritems():
                try:
                    task = project.get_task(t_id)
                    self.render_activity(project, task, resource, activities)
                except NodeNotFound:
                    print 'ignoring unknown', t_id
        self.drawer.open_line()
        self.drawer.main_title("")
        self.drawer.draw_timeline()
        self.drawer.close_line()

    def render_activity(self, project, task, resource, activities):
        """
        generate event for a resource activity
        """
        if self.options.del_ended and task.is_finished():
            log.info("  ignoring task", task.id, "for it's finished ",
                     " and using option --del-ended (or -D)")
            return
        if self.options.del_empty:
            begin = self.drawer._timeline_days[0]
            end = self.drawer._timeline_days[-1] + 1
            for a_begin, a_end, _, _, _, _ in activities:
                if (begin <= a_begin <= end) or (begin <= a_end <= end) :
                    break
            else :
                log.info("  ignoring task", task.id,
                         "for it has no activity within diagram bounds",
                         "and using option --del-empty (or -D)")
                return
        #else: option --keep-empty set. Draw all
        self.drawer.set_color_set(self._i)
        self._i += 1
        self.drawer.open_line()
        self.drawer.main_content(task.title, project, 2, task)
        self.drawer.activity_timeline(activities, resource)
        self.drawer.close_timeline()
        self.drawer.close_line()

    def render_total_occupation(self, project, resource):
        """
        generate event for global activity
        """
        log.info("Drawing occupation for", resource.name)
        self.drawer.set_color_set(self._i)
        self._i += 1
        self.drawer.open_line()
        self.drawer.main_content(resource.name)
        self.drawer.occupation_timeline(project, resource)
        self.drawer.close_timeline()
        self.drawer.close_line()
        
class ResourcesDrawer(AbstractDrawer) :
    """
    Concrete renderer which uses a std handler to draw a Resources diagram
    """
        
    def _get_table_dimension(self, project):
        """
        calculate dimension of the table
        """
        project.get_factor()
        self.factor=project.factor
        # calculate width
        width = TITLE_COLUMN_WIDTH
        if self.options.rappel:
            width *= 2
        if self.options.showids :
            width += FIELD_COLUMN_WIDTH
        if self.options.detail > 1 :
            width += FIELD_COLUMN_WIDTH*2
        if self.options.detail > 0 :
            width += FIELD_COLUMN_WIDTH*4
        width += len(self._timeline_days)*self._timestepwidth
        # calculate height
        nb_resources = len(project.resource_set.get_resources())
        nb_tasks = len(project.root_task.flatten())
        height = ROW_HEIGHT*(nb_tasks*nb_resources)
        return (width, height)
 
    # table heads #############################################################

    def legend(self):
        """ write the diagram's legend (calls private method)"""
        self._legend_task()
        self.open_line()
        self.close_line()
        self._legend_resource()
        self.close_line()
        
    def _legend_resource(self):
        """ write the diagram's legend of resources """
        self._draw_text('Resources Legend', style='italic', weight='bold')
        self._x += FIELD_COLUMN_WIDTH*3
        self._draw_rect(FIELD_COLUMN_WIDTH-10,
                  ROW_HEIGHT,
                  fillcolor=self._colors['EVEN_SET']['RESOURCE_UNAVAILABLE'])
        self._handler.draw_line(self._x + 1 , self._y + 1, 
                          self._x + FIELD_COLUMN_WIDTH - 10,
                          self._y + ROW_HEIGHT, 
                          color=self._colors['CONSTRAINT'])
        self._x += FIELD_COLUMN_WIDTH-10
        self._draw_text('unavailable', weight='bold')
        self._x += FIELD_COLUMN_WIDTH+10

    # project table content ###################################################

    def set_bg_color(self, day, available):
        """ set background color for a day"""
        if day.date == TODAY.date:
            bgcolor = self._color_set['TODAY']
        elif not available:
            bgcolor = self._colors['EVEN_SET']['RESOURCE_UNAVAILABLE']
        elif day.day_of_week in (5, 6):
            bgcolor = self._color_set['WEEKEND']
        else:
            bgcolor = self._color_set['WEEKDAY']
        return bgcolor

    def draw_separator_resources(self):
        if self.options.timestep == 'day':
            timestep = 1
        elif self.options.timestep == 'week':
            timestep = 7
        elif self.options.timestep == 'month':
            timestep = 31
        self._handler.draw_line(self._x, self._y, 
                                    self._x + FIELD_COLUMN_WIDTH/timestep,
                                    self._y, 
                                    color=(204, 204, 204))

    def occupation_timeline(self, project, resource):
        """
        write a timeline day for a resource occupation
        """
        for day in date_range(self.view_begin, self.view_end):
            # draw background
            self._handler.draw_rect(self._x, self._y, self._daywidth, ROW_HEIGHT,
                            fillcolor=self._colors['EVEN_SET']['RESOURCE_UNAVAILABLE'])
            for d in range(self.factor):
                day_ = day + oneHour * d*(HOURS_PER_DAY / self.factor)
                if day_.hour >= 12:
                    day_ += oneHour
                act = project.activities.select('resource', resource.id)
                if act:
                    act = act.select('begin', day_)
                usage = project.get_total_usage(resource.id, day_)
                available = resource.is_available(day_) and \
                               (day_ < self.view_end) and usage and act
                self._occupation_timeline(available, 1./self.factor, day_, d)
                if usage > 1:
                    log.info(" Warning! usage", usage, "for", resource.id, "on", day_)
            #draw separators
            self.draw_separator_resources()
            # update abscisse
            self._x += self._daywidth 

    def _occupation_timeline(self, available, usage, day, d):
        """
        write a day for a resource occupation
        """
        width = self._daywidth        
        height = 0
        color = None
        if not available:
            if usage:
                color = self._colors['TASK_SET']['problem']
                height = ROW_HEIGHT
        else:
            if usage > 1:
                color = self._colors['TASK_SET']['problem']
                height = ROW_HEIGHT
            elif 0 < usage <= 1:
                color = self._color_set['RESOURCE_USED']
                height = ROW_HEIGHT*min(usage, 1)
        if day.date == TODAY.date and not(color == self._color_set['RESOURCE_USED']):
            color = self._color_set['TODAY']
            height = ROW_HEIGHT
        if color and (available or color == self._color_set['TODAY']):
                self._handler.draw_rect(self._x+1, self._y+ROW_HEIGHT-height*(d+1),
                                width, height,
                                fillcolor=color)  # occupation totale(taches confondues)

    def activity_timeline(self, activities, resource):
        """write a day for a task"""
        task = activities[0][3]
        for day in date_range(self.view_begin, self.view_end):
            available = resource.is_available(day)
            bgcolor = self.set_bg_color(day, available)
            #draw backgrgound
            self._handler.draw_rect(self._x, self._y, self._daywidth, ROW_HEIGHT,
                                fillcolor=bgcolor)

            for d in range(self.factor):
                day_ = day + oneHour * d*(HOURS_PER_DAY / self.factor)
                if day_.hour >= 12:
                    day_ += oneHour
                for begin, end, _, _, usage, _ in activities:
                    if begin <= day_ < end:
                        if usage > 1:
                            log.info("    usage", usage,
                                     "for", resource.id, "on", day)
                            break
                        self._activity_timeline(available, usage, day, d)
            self._activity_timeline(False, 0, day, d)
            self.draw_separator_resources()
            # update abscisse
            self._x += self._daywidth

    def _activity_timeline(self, available, usage, day, d):
        """write a day for a task"""
        width = self._daywidth
        height = 0
        if 0 < usage <= 1:
            if available:
                color = self._color_set['RESOURCE_USED']
                height = ROW_HEIGHT*min(usage, 1)
            else:
                color = self._colors['TASK_SET']['problem']
                height = ROW_HEIGHT
        elif usage > 1:
            color = self._colors['TASK_SET']['problem']
            height = ROW_HEIGHT
        else:
            color = None

        if color:
            self._handler.draw_rect(self._x+1, self._y+ROW_HEIGHT-height*(d+1),
                                        width, height,
                                        fillcolor=color)




# -*- coding: iso-8859-1 -*-
# Copyright (c) 2004-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Projman - (c)2000-2005 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projects/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

__revision__ = "$Id: resource.py,v 1.4 2005-09-06 17:07:00 nico Exp $"


from logilab.common.tree import VNode 

class ResourceRole(VNode):
    """A resource_role i used to decribe skills of a resource, and the 
        hourly cost, depending of the skill
    
        attributes: 
            - type : qualifiacations (engineer, ...)
            - hourly cost: cost of one hour of work by the resource associated 
              to this resource_role
            
    """
    TYPE = 'resource_role'
    
    def __init__(self, id, name=u'', hourly_cost=0.0, unit='EUR'):
        VNode.__init__(self, id)
        self.id = id
        self.name = name
        self.hourly_cost = hourly_cost
        self.unit = unit

class ResourceRoleSet(VNode):

    TYPE = 'resource_role_set'

    def __init__(self, r_id):
        VNode.__init__(self, r_id)

    add_resource_role = VNode.append
    get_resource_role = VNode.get_node_by_id




# Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Abstract classes for Abstract classes for rendering """

__revision__ = "$Id: abstract.py,v 1.1 2005-09-06 17:07:22 nico Exp $"

from mx.DateTime import now
from projman import ENCODING
from projman.lib import date_range

TITLE_COLUMN_WIDTH  = 260
FIELD_COLUMN_WIDTH  = 40
TIME_COLUMN_WIDTH   = 40.
LEGEND_COLUMN_WIDTH = 80
ROW_HEIGHT          = 24

TODAY = now()

DEFAULT_COLORS = {
    'HEAD'       : 'lightgrey',
    'CONSTRAINT' : 'black',
    # for odd row
    'ODD_SET' : {'TITLE'  : 'white',
                 'FIELD'  : 'darkseagreen',
                 'WEEKDAY': 'wheat',
                 'WEEKEND': 'cornsilk',
                 'TODAY'  : 'salmon',
                 'RESOURCE_USED' : 'darkblue',
                 'RESOURCE_UNAVAILABLE' : 'cornsilk',
                 },
    # for even row
    'EVEN_SET' : {'TITLE'  : 'wheat',
                  'FIELD'  : 'teal',
                  'WEEKDAY': 'cornsilk',
                  'WEEKEND': 'wheat',
                  'TODAY'  : 'salmon',
                  'RESOURCE_USED' : 'darkblue',
                  'RESOURCE_UNAVAILABLE' : 'wheat',
                    },
    # task background according to its status
    'TASK_SET' : { 'todo'   : 'orange',   #'#7485f6',
                   'current': 'darkblue', #'#ffa703',
                   'done'   : 'darkgray', #'#818181',
                   'problem': 'red'       #'#ff0a4f',
                   }
    }

# FIXME use those colors in HTML renderer without css
def load_colors(path=None):
    """
    load the colors configuration from the given path
    if path is None, try to get the path from the PROJMAN_COLORS_FILE
    environment variable or from the ~/.projmanrc file
    """
    from os.path import expanduser, exists, join
    colors = DEFAULT_COLORS
    if path is None:
        from os import environ
        path = environ.get('PROJMAN_COLORS_FILE', None)
    if path is None:
        path = expanduser(join('~', '.projmanrc'))
    if exists(path):
        load_colors_from_stream(open(path), colors)
    return colors

def load_colors_from_stream(stream, colors=DEFAULT_COLORS):
    """
    load the colors configuration from a stream add set matching value in the
    <colors> dictionary
    """
    for line_num, line in enumerate(stream.readlines()):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        label, color = line.split('=')
        label, color = label.strip(), color.strip()
        if label[:4] == 'ODD_':
            colors_dict = colors['ODD_SET']
            key = label[4:]
        elif label[:5] == 'EVEN_':
            colors_dict = colors['EVEN_SET']
            key = label[5:]
        elif label[:5] == 'TASK_':
            colors_dict = colors['TASK_SET']
            key = label[5:].lower()
        else:
            colors_dict = colors
            key = label
        if not colors_dict.has_key(key):
            raise Exception('Unknown color identifier "%s" \
            in colors file line %s' % (label, line_num))
        colors_dict[key] = color
    return colors


class AbstractRenderer :
    """
    Implements methods common to all renderers.
    
    generates events according to the given options
    concrete renderer must implements corresponding hooks
    (see the IRender)
    """
    
    DEFAULT_OPTIONS = {'timestep' : 'day', 
                       'detail' : 2,
                       'depth' : 0,
                       'view_begin' : None,
                       'view_end' : None,
                       'showids' : False,
                       'rappel' : False
                       }
    
    def __init__(self, options) :
        """
        Initialise all fields and tools needed for rendering,
        BUT the Concrete Renderer still needs to define its drawer.
        """
        self.drawer = None
        self.options = options
        for k, v in self.DEFAULT_OPTIONS.items():
            if not hasattr( options, k ):
                setattr( options, k, v )

    def write(self, string) :
        """
        write a content part to the output buffer
        """
        self.output.write(string.encode(ENCODING))

    def render(self, project, stream) :
        """
        Method called to draw & return diagram. Must be overridden
        """
        self.output = stream
        self._render_init(project)
        self._render_body(project)
        self._render_end()
    
    def _render_init(self, project):
        """
        Method called in first place by render to set up display
        """
        # defines vars needed by _render_xxx
        self._pending_constraints = {}
        self._visible_tasks = {}
        self._i = 0 # row num
        # set borders
        begin, end = project.get_task_date_range(project.root_task)
        if self.options.view_begin:
            view_begin =  self.options.view_begin
        else:
            view_begin = begin
        if self.options.view_end:
            view_end = self.options.view_end
        else:
            view_end = end
        if not view_begin:
            view_begin = project.calculate_date_range()[0]
        if not view_end:
            view_end = project.calculate_date_range()[1]
        #open time line
        self.drawer.open_timeline(view_begin, view_end)
        #open table for display
        self.drawer.open_table(project)
            
    def _render_end(self):
        """
        Method called by render when it ends to finish display
        """
        # print table tail 
        self.drawer.open_line()
        self.drawer.main_title("Generated on %s" % (
            now().strftime("%Y/%m/%d %H:%M:%S")))
        self.drawer.draw_timeline()
        self.drawer.close_line()
        self.drawer.close_table()
        # print legend
        self.drawer.open_line()
        self.drawer.close_line()
        self.drawer.legend()
        # close
        self.drawer.close_drawing()

class AbstractDrawer :
    """
    Implements all hooks method called by the abstract renderer and do nothing
    """
    
    def __init__(self, options, handler, colors_file=None, colors_stream = None):
        """
        defines colors and initializes colors_set
        """
        # initialise colors
        self._colors = None
        if colors_stream is not None:
            self._colors = load_colors_from_stream(colors_stream)
        else:
            self._colors = load_colors(colors_file)
        self.set_color_set(0)
        # set last members
        self.options = options
        self._handler = handler
        self._timeline_days = []
        # ajust time step
        if self.options.timestep == 'day':
            self._timestep = 1
        if self.options.timestep == 'week':
            self._timestep = 7
        if self.options.timestep == 'month':
            self._timestep = 13   # maximum possible value
        # x and y coordinate for current element location
        self._x = 0
        self._y = 0        
        self._max_x = 0
        self._max_y = 0
        
    def set_color_set(self, parity):
        """
        Defines which color_set to use, EVEN or ODD form _colors
        """
        if parity % 2:
            self._color_set = self._colors['ODD_SET']
        else:
            self._color_set = self._colors['EVEN_SET']
    
    # low level methods ######################################################

    def _draw_text(self, text, **attrs):
        """
        draw a text with some alignement
        """
        self._handler.draw_text(self._x+3, self._y+18, text, **attrs)
        
    def _draw_rect(self, width, height, **attrs):
        """
        draw a rectangle with some alignement
        """
        self._handler.draw_rect(self._x+1, self._y+1, width-2, height-2, **attrs)

    def open_link(self,url):
        self._handler.open_link(url)

    def close_link(self):
        self._handler.close_link()

    # generic events ##########################################################

    def close_drawing(self): 
        self._handler.close_drawing( (0,0,self._max_x,self._max_y) )

    def open_table(self, project):
        """
        open the table where the resources diagram will be located
        """
        # adjust width according to timestep
        self._daywidth = TIME_COLUMN_WIDTH // self._timestep
        self._timestepwidth = self._daywidth * self._timestep
        self._init_table()

        # estimate drawing width/height
        width, height = self._get_table_dimension(project)
        self._handler.init_drawing(width, height)

        # number of optional fields
        self._fields = 0
        
    def _init_table(self):
        """
        abstract method used for specific initialisation
        """
    
    def close_table(self): 
        """
        close the table where the diagram is located
        """
        # nothing to be done, close line does all the work

    def open_line(self):
        """
        open a new line in the table
        """
        # nothing to be done, close line does all the work
        
    def close_line(self):
        """close a table's line"""
        self._max_x = max(self._max_x, self._x)
        self._max_y = max(self._max_y, self._y)
        self._x = 0
        self._y += ROW_HEIGHT

    def draw_timeline(self):
        """
        Format time line according to time step
        """
        for day in list(date_range(self.view_begin, self.view_end)):
            # timestep == week
            if self._timestep == 7 and day.day_of_week == 0:
                date = str(day.iso_week[1])
                self._draw_text("s%s" %date)
            #timestep == month
            elif self._timestep == 13 and day.day == 1:
                date = str("%02d" %day.month)
                self._draw_text(date)
            # timestep == day
            elif self._timestep == 1:
                date = day.strftime('%m-%d')
                self._draw_text(date)
            if self._timestep == 30:
                step = float(day.days_in_month)
            else:
                step = float(self._timestep)
            self._x += self._timestepwidth / step
        
    def open_timeline(self, begin, end):
        """
        open a new time line in the table
        """
        self._timeline_days = list(date_range(begin, end, self._timestep))
        self.view_begin = begin
        self.view_end = end

    def close_timeline(self):
        """
        close a table's time line
        """
        # nothing to be done
        
    # project table head/tail #################################################

    def main_title(self, title):
        """ write the main description column's title """
        self._draw_rect(TITLE_COLUMN_WIDTH,
                        ROW_HEIGHT, fillcolor=self._colors['HEAD'])
        self._draw_text(title, weight='bold')
        self._x += TITLE_COLUMN_WIDTH

    def simple_title(self, title, tail=False):
        """ write a simple description column's title """
        self._draw_rect(FIELD_COLUMN_WIDTH,
                        ROW_HEIGHT, fillcolor=self._colors['HEAD'])
        if not tail:
            self._fields += 1
            if title:
                self._draw_text(title, weight='bold')
        self._x += FIELD_COLUMN_WIDTH
        
    # standard table content ###################################################

    def simple_content(self, content):
        """ write a simple cell content  """
        self._draw_rect(FIELD_COLUMN_WIDTH, ROW_HEIGHT,
                        fillcolor=self._color_set['FIELD'])
        self._draw_text(content)
        self._x += FIELD_COLUMN_WIDTH

    def main_content(self, content, project=None,
                     depth=0, task=None):
        """ write a main cell's content """
        self._draw_rect(TITLE_COLUMN_WIDTH, ROW_HEIGHT,
                        fillcolor=self._color_set['TITLE'])
        if task is not None and task.TYPE != "milestone":
            try:
                status = project.get_task_status(task)
                color = self._colors['TASK_SET'][status]
            except KeyError, exc:
                print 'while drawing, ignoring', exc
                color = self._colors['TASK_SET']['problem']
            self._ctask = task
            self._color = color
        if depth:
            content = '%s%s'% ('  '*depth, content)
        self._draw_text(content, style='italic', weight='bold')
        self._x += TITLE_COLUMN_WIDTH

    def _legend_task(self) :
        """ write the diagram's legend of tasks """
        self._draw_text('Tasks Legend', style='italic', weight='bold')
        self._x += FIELD_COLUMN_WIDTH*2
        for status, color in self._colors['TASK_SET'].items():
            self._draw_rect(FIELD_COLUMN_WIDTH-10,
                            ROW_HEIGHT, fillcolor=color)
            self._x += FIELD_COLUMN_WIDTH-10
            self._draw_text(status, weight='bold')
            self._x += FIELD_COLUMN_WIDTH+10







import sys
import os
import os.path as osp
import gtk
import gtk.glade
import gobject

_main_module = sys.modules[__name__]
_main_dir = osp.dirname( _main_module.__file__)
_toplevel = osp.abspath(osp.join(_main_dir,"..",".."))

try:
    import projman
    import projman.projmanedit
except ImportError:
    if "projman" in sys.modules:
        del sys.modules["projman"]
    sys.path.insert(0,_toplevel) # for projman
    print "using path=", sys.path[0]
    import projman
    import projman.projmanedit

from projman.readers import ProjectXMLReader
from projman.writers.projman_writer import write_tasks_as_xml
from projman.projmanedit.gui.taskedit import TaskEditor

GLADE=projman.projmanedit.GLADE

XMLFILTER = gtk.FileFilter()
XMLFILTER.set_name("XML file")
XMLFILTER.add_pattern("*.xml")
ANYFILTER = gtk.FileFilter()
ANYFILTER.set_name("Any file")
ANYFILTER.add_pattern("*")

import gtksourceview2


def glade_custom_handler(glade, function_name, widget_name,
			str1, str2, int1, int2):
    if function_name == "create_sourceview":
        widget = gtksourceview2.View()
    else:
        debug("Unknown custom widget : %s/%s", function_name, widget_name )
        widget = None
    if widget:
        widget.show()
    return widget

gtk.glade.set_custom_handler( glade_custom_handler )

class MainApp(gobject.GObject):
    """The main application
    """
    __gsignals__ = {'project-changed': (gobject.SIGNAL_RUN_FIRST,
                                        gobject.TYPE_NONE,
                                        () ),
                    }

    def __init__(self):
        gobject.GObject.__init__(self)
        self.project_file = None
        self.project = None
        self.files = None
        self.ui = gtk.glade.XML(GLADE,"window_main")
        self.ui.signal_autoconnect(self)
        # build specific ui controlers
        self.taskeditor = TaskEditor( self )

    def get_project_path(self):
        return self.project_file[0:self.project_file.rindex("/")+1]

    def on_new_cmd_activate(self,*args):
        print "new", args

    def on_open_cmd_activate(self,*args):
        print "open", args
        dlg = gtk.FileChooserDialog(title=u"Open project",
                                    action=gtk.FILE_CHOOSER_ACTION_OPEN,)
        dlg.add_button( gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL )
        dlg.add_button( gtk.STOCK_OK, gtk.RESPONSE_OK )
        dlg.add_filter( XMLFILTER )
        dlg.add_filter( ANYFILTER )
        res = dlg.run()
        fname = dlg.get_filename()
        dlg.destroy()
        if res!=gtk.RESPONSE_OK:
            return
        self.load_project( fname )

    def load_project(self, fname):
        self.project_file = fname
        reader = ProjectXMLReader( fname, None, True )
        self.project, self.files = reader.read()
        from projman.scheduling.csp import CSPScheduler
        scheduler = CSPScheduler(self.project)
        scheduler.schedule()
        self.project = scheduler.project
        from projman.writers.projman_writer import write_schedule_as_xml
        schedule_file=str(self.get_project_path())+str(self.files["schedule"])
        write_schedule_as_xml(self.get_project_path()+self.files["schedule"],self.project)

        from projman.renderers import GanttRenderer, HandlerFactory
        handler = HandlerFactory("svg")
        # it works !! but HOW  ??? ......
        options = handler
        options.timestep = 1
        options.detail = 2
        options.depth = 0
        options.view_begin = None
        options.view_end = None
        options.showids = False
        options.rappel = False
        options.output = None
        options.selected_resource = None
        options.format = "svg" # ce n est plus le format par defaut ...
        options.del_ended = False
        options.del_empty = False
        # end of mystic code ...
        renderer = GanttRenderer(options, handler)
        output = self.get_project_path()+"gantt.svg"
        stream = handler.get_output(output)
        renderer.render(self.project, stream)
        self.taskeditor.w("image1").set_from_file(output)
        self.emit("project-changed")
        
    def on_save_cmd_activate(self,*args):
        print "save", args
        basedir = osp.dirname( self.project_file )
        task_file = osp.join( basedir, self.files['tasks'] )
        write_tasks_as_xml( task_file, self.project )
        self.load_project(self.project_file)


    def on_save_as_cmd_activate(self,*args):
        print "save as", args
        dlg = gtk.FileChooserDialog('Save as...',
                                    None,
                                    gtk.FILE_CHOOSER_ACTION_SAVE,
                                    (gtk.STOCK_CANCEL,
                                     gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_SAVE,
                                     gtk.RESPONSE_OK))
        dlg.add_filter( XMLFILTER )
        dlg.add_filter( ANYFILTER )
        res = dlg.run()
        fname = dlg.get_filename()
        dlg.destroy()
        if res!=gtk.RESPONSE_OK:
            return
        write_tasks_as_xml( fname, self.project )
        self.load_project(self.project_file)

    def on_quit_cmd_activate(self, *args):
        gtk.main_quit()

    def on_window_main_destroy(self, *args):
        gtk.main_quit()


app = MainApp()

if len(sys.argv)>1:
    app.load_project( sys.argv[1] )

gtk.main()









from gcsp import ProjmanProblem, solve

DURATIONS = [ 20, 60, 20, 20, 20, 30 ]


def reg_csp( func, l0, l1 ):
    for p0 in l0:
        for p1 in l1:
            func( p0, p1 )

def problem( max_duration ):
    NW0 = [ i for i in range(max_duration) if i%7 in (5,6) ]
    NW1 = [ i for i in range(max_duration) if i%7 in (4,5,6) ]
    pb = ProjmanProblem( 6, 3, max_duration )
    for i,d in enumerate(DURATIONS):
        pb.set_duration( i, d )
    T0 = []
    T1 = []
    T2 = []
    T3 = []
    T4 = []
    T5 = []
    T0.append( pb.alloc( 0, 1 ) )
    T1.append( pb.alloc( 1, 0 ) )
    T1.append( pb.alloc( 1, 1 ) )
    T1.append( pb.alloc( 1, 2 ) )
    T2.append( pb.alloc( 2, 0 ) )
    T3.append( pb.alloc( 3, 1 ) )
    T4.append( pb.alloc( 4, 1 ) )
    T4.append( pb.alloc( 4, 2 ) )
    T5.append( pb.alloc( 5, 0 ) )
    T5.append( pb.alloc( 5, 1 ) )

    pb.begin_after_end( 0, 1 )
    pb.begin_after_end( 0, 2 )
    pb.begin_after_end( 0, 3 )
    pb.begin_after_end( 1, 4 )
    pb.begin_after_end( 2, 4 )
    pb.begin_after_end( 3, 4 )
    pb.begin_after_begin( 4, 5 )
    

    for nw in NW0:
        pb.add_not_working_day( 0, nw )
        pb.add_not_working_day( 2, nw )
    for nw in NW1:
        pb.add_not_working_day( 1, nw )

    return pb
import sys
pb = problem( int(sys.argv[1]) )
solve( pb )




# Copyright (c) 2000-2004 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Projman - (c)2004 Logilab - All rights reserved."""

import shutil, os, tempfile
import os.path as osp

from logilab.common import testlib
from logilab.common.clcommands import cmd_run

from projman import commands # load commands

from projman.test import DATADIR, XML_PROJMAN, XML_SCHEDULED_PROJMAN, \
     XML_SCHEDULED_PROJMAN_FULL, TAR_PROJMAN


XML_PROJMAN = osp.join(DATADIR, XML_PROJMAN)
XML_SCHEDULED_PROJMAN = osp.join(DATADIR, XML_SCHEDULED_PROJMAN)
XML_SCHEDULED_PROJMAN_FULL = osp.join(DATADIR, XML_SCHEDULED_PROJMAN_FULL)
TAR_PROJMAN = osp.join(DATADIR, TAR_PROJMAN)

class AbstractCommandTest(testlib.TestCase):
    """testing """
    
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
    
    def tearDown(self):
        shutil.rmtree(self.tmpdir) 
    
class ScheduleTest(AbstractCommandTest):
    """testing """

    def setUp(self):
        AbstractCommandTest.setUp(self)
        self.projman_path =  osp.join(DATADIR, "tmp_projman.xml")
        shutil.copyfile(XML_PROJMAN, self.projman_path)
        self.sched = osp.join(DATADIR, 'schedule.xml')

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        os.remove(self.projman_path)

    def test_default(self):
        """Check if a schedule file is created and can be read again"""
        self.skip("this test needs rewrite")
        try:
            os.remove(self.sched)
        except OSError, e:
            print "ERROR:", self.sched, e
            pass
        print "PROJECT FILE:", self.projman_path
        cmd_run('schedule', '--type', 'csp', '-f', self.projman_path)
        #self.assert_(osp.exists(self.sched))
        # try reschedule
        cmd_run('schedule', '--type', 'csp', '-f', self.projman_path)
        #self.assert_(osp.exists(self.sched))

class DiagramTest(AbstractCommandTest):
    
    def test_gantt(self):
        gantt = osp.join(self.tmpdir, 'out_gantt.svg')
        cmd_run('diagram', '--timestep', 'week', '-f', XML_SCHEDULED_PROJMAN, 'gantt')
        self.assert_(osp.exists('gantt.svg'))
        os.remove('gantt.svg')
        cmd_run('diagram', '--timestep', 'week', '-f', XML_SCHEDULED_PROJMAN, '-o',
                gantt, 'gantt')
        self.assert_(osp.exists(gantt))
        
#    def test_gantt2(self):
#        gantt = osp.join(self.tmpdir, 'out_gantt.svg')
#        cmd_run('diagram', '--timestep', 'week',
#                '-f', XML_SCHEDULED_PROJMAN_FULL, 'gantt')
#        self.assert_(osp.exists("gantt.svg"))
#        os.remove('gantt.svg')
#        cmd_run('diagram', '--timestep', 'week', '-f', XML_SCHEDULED_PROJMAN_FULL,
#                '-o', gantt, 'gantt')
#        self.assert_(osp.exists(gantt))

    def test_resources(self):
        resources = osp.join(self.tmpdir, 'resources')
        cmd_run('diagram', '-f', XML_SCHEDULED_PROJMAN, 'resources',
                '-o', resources)
        self.assert_(osp.exists(resources))
        cmd_run('diagram', '-f', XML_SCHEDULED_PROJMAN, '--format', 'svg',
                'resources', '-o', resources)
        self.assert_(osp.exists(resources))
    
    def test_gantt_resources(self):
        img = osp.join(self.tmpdir, 'gantt-resources')
        cmd_run('diagram', '-f', XML_SCHEDULED_PROJMAN, 'gantt-resources',
                '-o', img,)
        self.assert_(osp.exists(img))


class XmlTest(AbstractCommandTest):
    
    def test_all(self):
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN,
                'duration-table', 'duration-section', 'tasks-list-section',
                'cost-para', 'cost-table', 'rates-section')
        self.assert_(osp.exists("output.xml"))
        
    def test_duration_table(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN, '-o', out, 'duration-table')
        self.assert_(osp.exists(out))
        
    def test_duration_section(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN, '-o', out, 'duration-section')
        self.assert_(osp.exists(out))
    
    def test_tasks_list_section(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN,
                '-o', out, 'tasks-list-section')
        self.assert_(osp.exists(out))

    def test_cost_para(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN, '-o', out, 'cost-para')
        self.assert_(osp.exists(out))

    def test_cost_table(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN, '-o', out, 'cost-table')
        self.assert_(osp.exists(out))

    def test_rates_section(self):
        out = osp.join(self.tmpdir, 'out_duration.xml')
        cmd_run('view', '-f', XML_SCHEDULED_PROJMAN, '-o', out, 'rates-section')
        self.assert_(osp.exists(out))

if __name__ == '__main__':
    testlib.unittest_main()



# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
Projman - (c)2000-2005 LOGILAB <contact@logilab.fr> - All rights reserved.
Home: http://www.logilab.org/projman

Render a Project to HTML.
"""

__revision__ = "$Id: html.py,v 1.1 2005-09-06 17:07:26 nico Exp $"

from projman.lib import *
from projman.renderers.resource import *

HTML_CSS = {
    'project':           ('class="project"','',''),
    'timeline':          ('class="timeline"','',''),
    'resource':          ('class="resource"','',''),
    'holiday':           ('class="holiday"','',''),
    'workday':           ('class="workday"','',''),
    'today':             ('class="today"','',''),
    'calendar':          ('class="calendar"','',''),
    'cont_todo':         ('class="cont_todo"','',''),
    'cont_ready':        ('class="cont_ready"','',''),
    'cont_run':          ('class="cont_run"','',''),
    'cont_done':         ('class="cont_done"','',''),
    'leaf_todo':         ('class="leaf_todo"','',''),
    'leaf_ready':        ('class="leaf_ready"','',''),
    'leaf_run':          ('class="leaf_run"','',''),
    'leaf_done':         ('class="leaf_done"','',''),
    'leaf_problem':      ('class="leaf_problem"','',''),
    'leaf_date_todo':    ('class="leaf_date_todo"','',''),
    'leaf_date_ready':   ('class="leaf_date_ready"','',''),
    'leaf_date_run':     ('class="leaf_date_run"','',''),
    'leaf_date_done':    ('class="leaf_date_done"','',''),
    'leaf_date_problem': ('class="leaf_date_problem"','',''),
    'cont_date_todo':    ('class="cont_date_todo"','',''),
    'cont_date_ready':   ('class="cont_date_ready"','',''),
    'cont_date_run':     ('class="cont_date_run"','',''),
    'cont_date_done':    ('class="cont_date_done"','',''),
    'cont_date_problem': ('class="cont_date_problem"','',''),
    }

COLOR_TASK = { 'todo' :  ("6666FF", "AAAAFF", "9999FF"),
               'ready':  ("66FF66", "AAFFAA", "99FF99"),
               'run'  :  ("FF8800", "FFBB44", "FFBB33"),
               'done' :  ("666666", "AAAAAA", "BBBBBB"),
               'problem':("FF0000", "FF8888", "")
               }
HTML_NOCSS = {
    'project':           ('','<font size="3" color="orangered">','</font>'),
    'timeline':          ('','',''),
    'resource':          ('','<b>','</b>'),
    'holiday':           ('bgcolor="#DDDDDD"','',''),
    'workday':           ('bgcolor="#%s"'%COLOR_TASK['todo'][0],'',''),
    'today':             ('', '<font size="1">','</font>'),
    'calendar':          ('','',''),
    'cont_todo':         ('bgcolor="#%s"'%COLOR_TASK['todo'][2],'<i>','</i>'),
    'cont_ready':        ('bgcolor="#%s"'%COLOR_TASK['ready'][2],'<i>','</i>'),
    'cont_run':          ('bgcolor="#%s"'%COLOR_TASK['run'][2],'<i>','</i>'),
    'cont_done':         ('bgcolor="#%s"'%COLOR_TASK['done'][2],'<i>','</i>'),
    'leaf_todo':         ('bgcolor="#%s"'%COLOR_TASK['todo'][2],'',''),
    'leaf_ready':        ('bgcolor="#%s"'%COLOR_TASK['ready'][2],'',''),
    'leaf_run':          ('bgcolor="#%s"'%COLOR_TASK['run'][2],'',''),
    'leaf_done':         ('bgcolor="#%s"'%COLOR_TASK['done'][2],'',''),
    'leaf_problem':      ('bgcolor="#%s"'%COLOR_TASK['problem'][2],'',''),
    'leaf_date_todo':    ('bgcolor="#%s"'%COLOR_TASK['todo'][0],'',''),
    'leaf_date_ready':   ('bgcolor="#%s"'%COLOR_TASK['ready'][0],'',''),
    'leaf_date_run':     ('bgcolor="#%s"'%COLOR_TASK['run'][0],'',''),
    'leaf_date_done':    ('bgcolor="#%s"'%COLOR_TASK['done'][0],'',''),
    'leaf_date_problem': ('bgcolor="#%s"'%COLOR_TASK['problem'][0],'',''),
    'cont_date_todo':    ('','',''),
    'cont_date_ready':   ('','',''),
    'cont_date_run':     ('','',''),
    'cont_date_done':    ('','',''),
    'cont_date_problem': ('','',''),
    }


class ResourcesHTMLRenderer(ResourcesRenderer) :
    """render a project to a gantt HTML diagram"""
    
    def __init__(self, options, css=None):
        ResourcesRenderer.__init__(self, options)
        if css:
            self._ref = HTML_CSS
        else:
            self._ref = HTML_NOCSS

    def open_table(self, activities):
        """ open the table where the resources diagram will be located """
        self._resource = None
        self.write("<table border='0'>\n")
        
    def close_table(self):
        """ close the table where the resources diagram is located """
        self.write("</table>\n")

    def open_line(self):
        """ open a new line in the table """
        self.write("<tr>")
    def close_line(self):
        """ close a table's line """
        self.write("</tr>")
        

    # project heads ######################################################
    def main_title(self, title):
        """ write the main description column's title """
        self.write("<th colspan='2' %s>%s %s%s</th>\n" % (
            self._ref['project'][0], self._ref['project'][1],
            title, self._ref['project'][2]))
        
    def simple_title(self, title, tail=0):
        """ write a simple description column's title """
        self.write("<th>%s</th>" % (title or '&nbsp;'))
            
    def timeline_title(self, date, today=0):
        """ write a timeline column title """
        if today:
            classe = 'today'                
        else :
            classe = 'timeline'
        self.write("<th %s>%s%s%s</th>" % (self._ref[classe][0],
                                           self._ref[classe][1],
                                           date, self._ref[classe][2]))

    def legend(self) :
        """ write the diagram's legend """
        self.write("<table cellpadding='3'><tr><td>Legend</td>")
        for t in ('todo', 'ready', 'run', 'problem'):
            attrs = self._ref['leaf_date_%s'%t]
            self.write("<td %s>%s%s%s</td>" % (attrs[0], attrs[1], t, attrs[2]))
        self.write("</tr></table>")
        
    # project table content ###################################################
    def main_content(self, content, depth=0, activities=None):
        """ write a main cell's content """
        if activities is not None:
            self._colspan = 0
            self.open_line()
            classe = 'project'
        else:
            classe = status_class = 'resource'
        self._colspan = 0    
        self._attrs = attrs = self._ref[classe]
        self.write("<tr valign='top'><td %s>%s&nbsp;%s</td><td %s>%s%s%s</td>" \
                   % (self._ref[status_class][0], self._ref[status_class][1],
                      self._ref[status_class][2], attrs[0],
                      attrs[1], content or "&nbsp;", attrs[2]))

    def simple_content(self, content):
        """ write a simple cell content """
        attrs = self._attrs
        self.write("<td %s>%s%s%s</td>" % (attrs[0],
                                           attrs[1], content or '&nbsp;',
                                           attrs[2]))

    def close_timeline(self):
        """ close a table's time line """
        if self._resource is not None:
            self.flush_resource_cells()
        else:
            self.flush_task_cells('')

    def flush_task_cells(self, text):
        """ flush pending task timeline cells """
        if self._colspan:
            _attrs = self._status_attrs
            self.write("<td colspan='%s' %s>%s%s%s</td>" % (
                self._colspan, _attrs[0], _attrs[1],
                (text or '&nbsp;'), _attrs[2]))
        
    # resources table content ##################################################

    def flush_resource_cells(self):
        """ flush pending resource timeline cells """
        workday = self._ref['workday']
        if self._colspan:
            self.write("<td colspan='%s' %s>%s%s</td>" % (
                self._colspan, workday[0], workday[1],workday[2]))
            self._colspan = 0

    def resource_timeline(self, usage, worked, text, day, resource):
        """ write a timeline day for a resource """
        self._resource = 1
        if not worked:
            self.flush_resource_cells()
            attrs = self._ref['holiday']
            self.write("<td %s>%s%s%s</td>" % (attrs[0], attrs[1], text, attrs[2]))
        elif not usage:
            self.flush_resource_cells()
            attrs = self._ref['calendar']
            self.write("<td %s>%s%s%s</td>" % (attrs[0], attrs[1], text, attrs[2]))
        else :
            self._colspan += 1



# -*- coding: iso-8859-1 -*-
# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
schedule project without caring about resource conflicts
"""

from mx.DateTime import today
from projman.lib.constants import *

import logging
log = logging.getLogger("simple")

def cmp_tasks(task_a, task_b):
    """
    a < b return -1
    a == b return 0
    a > b return 1
    """
    # at least one task is None
    if task_a is None and task_b is None:
        return 0
    if task_a is None :
        return -1
    if task_b is None:
        return 1
    # otherwise
    res = 0
    family_a = [task.id for task in task_a.lineage()]
    family_b = [task.id for task in task_b.lineage()]
    for constraint_type, task_id in task_a.get_task_constraints():
        if constraint_type == BEGIN_AFTER_END:
            if task_id in family_b:
                res = 1
                break
    for constraint_type, task_id in task_b.get_task_constraints():
        if constraint_type == BEGIN_AFTER_END:
            if task_id in family_a:
                if res:
                    raise Exception("a after b and b after a ?!?")
                res = -1
                break
    if res == 0:
        if task_a.priority is None and task_b.priority is None:
            res = 0
        elif task_a.priority is None:
            res = 1
        elif task_b.priority is None:
            res = -1
        else:
            res = cmp(task_a.priority, task_b.priority)
    log.debug('%s %s %s'%(task_a, {-1:'<',0:'=',1:'>'}[res], task_b))
    return res

class SimpleScheduler:
    """
    schedule simple projects by ordering tasks and splitting resources
    """
    def __init__(self, project):
        self.project = project
        self.date_res = {}

    def earliest_begin(self, node):
        begin = None
        for constraint_type, task_id in node.get_task_constraints():
            if constraint_type == BEGIN_AFTER_END:
                b,e = self.project.get_task_date_range(self.project.get_task(task_id))
                begin = max(begin or e, e)
        for c_type, date in node.get_date_constraints():
            if c_type == BEGIN_AFTER_DATE :
                begin = max(begin or date, date)
        return begin
        
    def _process_node(self, node):
        # get dates
        begin = self.earliest_begin(node) or today()
        duration = max(0, node.duration-1)
        #if begin and not end:
        #    end = begin + duration
        #elif end and not begin:
        #    begin = end - duration
        # add activities
        activities = []
        resources = [(self.project.get_resource(r_id), usage/100)
                     for r_type, r_id, usage in node.get_resource_constraints()]
        if not resources:
            log.warning("task %s has no resource and will not be scheduled",node.id)
            return
        self.date_res = {}
        for res, usage in resources :
            self.date_res.setdefault(res.id, begin-1)
        load = node.duration*(1-node.progress)
        date = begin
        if load == 0 and begin == today():
            log.warning("task %s is complete but was never worked on"%node.id)
        while load > 0 and resources :
            for resource, usage in resources:
                if self.date_res[resource.id] < date and resource.is_available(date) \
                   and self.project.get_total_usage(resource.id, date) <= 1-usage :
                    activities.append( (date, date, resource.id, node.id, usage) )
                    self.date_res[resource.id] = date
                    load -= usage
                    if load <= 0:
                        break
            date += 1
        self.project.add_schedule(activities)

    def get_ordered_buckets(self):
        tasks = self.project.root_task.leaves()
        tasks.sort(cmp_tasks)
        buckets = []
        bucket = [None]
        for task in tasks:
            for other in bucket:
                if cmp_tasks(task,other) == 1:
                    bucket = [task]
                    buckets.append(bucket)
                    break
            else:
                bucket.append(task)
        return buckets

    def schedule(self, verbose=0):
        """
        Update the project's schedule
        Return list of errors occured during schedule
        """
        activities = []
        for bucket in self.get_ordered_buckets():
            for leaf in bucket:
                self._process_node(leaf)
        return []




# -*- coding: iso-8859-1 -*-
"""
unit tests for module projman.lib.Project

Projman - (c)2000-2002 LOGILAB <contact@logilab.fr> - All rights reserved.
Home: http://www.logilab.org/projman

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

from mx.DateTime import now

from logilab.common import testlib

from projman.lib import MileStone, Task, Project, ResourcesSet, \
     Resource

class ProjectTC(testlib.TestCase):

    def setUp(self):
        """called before each test from this class

        parent (inge1/1, inge2/1)
          |
          |--- child1 (inge1/1), 10
          |
          |--- child2
          |      |
          |      |--- child2_1 (inge1/0.6), 5 
          |      |
          |      `--- child2_2 (inge2/1), 12
          |
          `--- stone [milestone]
        """
        self.parent = Task('parent')
        self.child1 = Task('child1')
        self.child2 = Task('child2')
        self.child2_1 = Task('child2_1')
        self.child2_2 = Task('child2_2')
        self.stone = MileStone('stone')
        # building tree
        self.parent.append(self.child1)
        self.parent.append(self.child2)
        self.child2.append(self.child2_1)
        self.child2.append(self.child2_2)
        self.parent.append(self.stone)
        # set duration
        self.child1.duration = 10
        self.child2_1.duration = 5
        self.child2_2.duration = 12
        # set of resources constraints
        self.parent.add_resource_constraint('worker', 'inge1')
        self.parent.add_resource_constraint('worker', 'inge2')
        self.child1.add_resource_constraint('worker', 'inge1')
        self.child2_1.add_resource_constraint('worker', 'inge1')
        self.child2_2.add_resource_constraint('worker', 'inge2')
        # create project
        self.project = Project()
        self.project.root_task = self.parent


    def test_project(self):
        self.assertEquals(self.project.root_task, self.parent)
        self.assertEquals(self.project.get_task('child2_2'), self.child2_2)

    def test_project_errors(self):
        """tests inconsistent projects"""
        # expliclity add duplicate node id
        self.parent.append(Task('child1'))
        self.project.root_task = self.parent
        self.assertEquals(self.project.check_consistency(), ['Duplicate task id: child1'])

    def _test_schedule(self):
        pass

    def test_resources(self):
        # build resources
        jean = Resource('inge1', 'Jean')
        paul = Resource('inge2', 'Paul')
        resource_set = ResourcesSet('rset')
        resource_set.append(jean)
        resource_set.append(paul)
        self.project.resource_set = resource_set
        # test setters & getters
        self.assertEquals(self.project.has_resource('inge1'), True)
        self.assertEquals(self.project.has_resource('Jean'), False)
        self.assertEquals(self.project.get_resource('inge1'), jean)
        self.assertEquals(self.project.get_resources(), ['inge1', 'inge2'])
    
    def test_tasks(self):
        self.assertEquals(self.project.get_task('stone'), self.stone)
        self.assertEquals(self.project.get_nb_tasks(), 6)
        self.assertEquals(self.project.is_in_allocated_time('child1', now()), False)

    def _test_activities(self):
        pass
    
    def _test_usage(self):
        pass
    
    def _test_total_usage(self):
        pass

    def _test_resources_disponibility(self):
        pass

    def _test_task_date_range(self):
        pass
    
    def _test_duration_worked_for(self):
        pass

    def _test_task_cost(self):
        pass

    

if __name__ == '__main__':
    testlib.unittest_main()




# Copyright (c) 2000-2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Exceptions used in the base model
"""

__revision__ = "$Id: _exceptions.py,v 1.2 2005-06-14 15:22:25 arthur Exp $"

from logilab.common.tree import NodeNotFound


class ProjectValidationError(Exception):
    """ base class for malformed project exceptions """

class MissingRequiredAttribute(ProjectValidationError):
    """ a required attribute is missing """
EX_MISSING_RATTR = 'file %s line %s : Missing required attribute %s to element <%s>'

class MalformedProjectFile(ProjectValidationError):
    """ malformed xml project description """
EX_MALFORMED_FILE = 'file %s line %s : <%s> should not be a child of <%s>'

class MalformedId(ProjectValidationError):
    """ non ascii id attribute"""

class ScheduleCycle (ProjectValidationError):
    """ constraints cycle exception """
EX_CYCLE_DETECTED = "Schedule cycle detected for task '%s'"

class ResourceNotFound (NodeNotFound):
    """ Unable to find resources exception """
EX_RESOURCE_BY_ID_NOT_FOUND = 'Resource id="%s" not found'
EX_RESOURCE_BY_SKILLS_NOT_FOUND = 'No resource with skill(s) "%s" not found'

class DuplicatedResource(Exception):
    """ Duplicated resource id exception """
EX_DUPLICATED_RESOURCE = 'A resource with id="%s" already exists'

class DuplicatedTaskId(Exception):
    """ Duplicated task id exception """
EX_DUPLICATED_TASK = 'A resource with id="%s" already exists'

class TTException(Exception):
    """ a TimeTable exception """

class ViewException(Exception):
    """ too much sub task to generate views """



# Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
a PIL handler for the Renderer
"""

import os
from os.path import exists
from cStringIO import StringIO
from PIL import Image, ImageDraw, ImageFont
import projman

__revision__ = "$ld $"

COLORS = {
    'aqua':         (0x00, 0xFF, 0xFF),
    'black':        0,    
    'blue':         (0x00, 0x00, 0xFF),
    'blueviolet':   (0x8A, 0x2B, 0xE2),
    'brown':        (0xA5, 0x2A, 0x2A),
    'cadetblue':    (0x5F, 0x9E, 0xA0),
    'chartreuse':   (0x7F, 0xFF, 0x00),
    'cornsilk':     (0xFF, 0xF8, 0xDC),
    'darkblue':     (0x00, 0x00, 0x8B),
    'darkgoldenrod':(0xB8, 0x86, 0x0B),
    'darkgray':     (0xA9, 0xA9, 0xA9),
    'darkgreen':    (0x00, 0x64, 0x00),
    'darkmagenta':  (0x8B, 0x00, 0x8B),
    'darkorange':   (0xFF, 0x8C, 0x00),
    'darkseagreen': (0x8F, 0xBC, 0x8F),
    'fuchsia':      (0xFF, 0x00, 0xFF),
    'green':        (0x00, 0x80, 0x00),
    'greenyellow':  (0xAD, 0xFF, 0x2F),
    'indigo':       (0x4B, 0x00, 0x82),
    'lightpink':    (0xFF, 0xB6, 0xC1),
    'lightgrey':    (0xD3, 0xD3, 0xD3),
    'limegreen':    (0x00, 0xFF, 0x00),
    'magenta':      (0xFF, 0x00, 0xFF),
    'mediumorchid': (0xBA, 0x55, 0xD3),
    'olive':        (0x80, 0x80, 0x00),
    'orange':       (0xFF, 0xA5, 0x00),
    'purple' :      (0x80, 0x00, 0x80),
    'red':          (0xFF, 0x00, 0x00),
    'salmon':       (0xFA, 0x80, 0x72),
    'teal' :        (0x00, 0x80, 0x80),
    'violet':       (0xEE, 0x82, 0xEE),
    'wheat':        (0xF5, 0xDE, 0xB3),
    'white':        (0xFF, 0xFF, 0xFF),
    'yellow':       (0xFF, 0xFF, 0x00),
    }

def delta_color(color, x):
    """ increment or decrements color by x """
    if color == 0:
        color = (0, 0, 0)
    if color[0] == '#':
        return (eval('0x%s'% color[1:3])+x,
                eval('0x%s'% color[3:5])+x,
                eval('0x%s'% color[5:7])+x)
    if len(color) == 3 and isinstance(color[0],int):
        one, two, three = color
        return (one+x, two+x, three+x)
    one, two, three = COLORS[color]
    return (one+x, two+x, three+x)


def rgb(color):
    """ return a triplet representing a RGB color """
    if color[0] == '#':
        return (eval('0x%s'%color[1:3]),
                eval('0x%s'%color[3:5]),
                eval('0x%s'%color[5:7]))
    return COLORS[color]

DEFAULT_FONT_DIRS = ('/usr/share/projman',
                     os.path.join(projman.__path__[0],
                                  "fonts"))

for font_dir in DEFAULT_FONT_DIRS:
    if exists(font_dir):
        FONT_DIR = font_dir
        break
else:
    FONT_DIR = None

FORMATS = ("ARG", "BMP", "CUR", "DCX", "EPS", "FLI", "FPX", "GBR", 
           "GIF", "ICO", "IM", "IPTC", "JPEG", "MIC", "MPEG", "MSP",
           "PCD", "PCX", "PDF", "PNG", "PPM", "PSD", "SGI", "SUN", 
           "TGA", "TIFF", "WMF", "XBM", "XPM")

def load_font(weight='', style=''):
    """ load font with the given weight and style """
    if style:
        style = ' %s' % style.capitalize()
    if weight:
        weight = ' %s' % weight.capitalize()
    font_name = 'Arial%s%s_12_72.pil' % (weight, style)
    #font_name = 'Courier Regular%s%s_12_72.pil' % (weight, style)
    #font_name = 'Times New Roman%s%s_12_72.pil' % (weight, style)
    #import sys
    #print >>sys.stderr, FONT_DIR, font_name
    return ImageFont.load(os.path.join(FONT_DIR, font_name))

class PILHandler:
    """ handler for the Python Imaging Library """
    
    def __init__(self, format):
        format = format.upper()
        assert format in FORMATS, 'Unknown format %s' % format
        self._format = format
        
    def init_drawing(self, width, height):
        """ initialize a new picture """
        #print 'diagram pixel size', width, height
        self._im = Image.new('RGB', (width, height), 0xFFFFFF)
        self._d = ImageDraw.Draw(self._im)
        self._default_font = load_font()
        
    def close_drawing(self, cropbox=None):
        """ close the current picture """
        if cropbox:
            self._im = self._im.crop(cropbox)
        # print 'diagram pixel size', width, height
    
    def get_result(self):
        """ return the current picture """
        print 'Deprecated'
        _buffer = StringIO()
        self._im.save(_buffer, self._format)
        return _buffer.getvalue()
        
    def save_result(self, stream):
        """ return the current picture """
        self._im.save(stream, self._format)

    def draw_text(self, x, y, text, **args):
        """ draw a text """
        if not args.has_key('color'):
            args['color'] = 'black'
        attrs = self._get_attrs(args)
        self._d.text((x, y-12), text.encode(projman.ENCODING), \
                     font=self._font, **attrs)
        
    def draw_line(self, x1, y1, x2, y2, **args):
        """ draw a line """
        attrs = self._get_attrs(args)
        self._d.line((x1, y1, x2, y2), **attrs)#fill=128)
        
    def draw_rect(self, x, y, width, height, **args):
        """ draw a rectangle """
        attrs = self._get_attrs(args)
        self._d.rectangle((x, y, x+width, y+height), **attrs)
        
    def draw_poly(self, point_list, **args):
        """ draw a polygon """
        attrs = self._get_attrs(args)
        self._d.polygon(point_list, **attrs)
        
    def _get_attrs(self, args):
        """ convert standard styles arguments to PIL attributes"""
        attrs = {}
        style, weight = '', ''
        for arg, value in args.items():
            if arg == 'style':
                style = value
            elif arg == 'weight':
                weight = value
            elif arg == 'fillcolor':
                attrs['fill'] = rgb(value)
            elif arg == 'color':
                attrs['fill'] = rgb(value)
            elif arg == 'delta_color':
                attrs['fill'] = delta_color(attrs['fill'], int(value))
            else:
                raise Exception('Unkwnown arguments %s'%arg)
            if style or weight:
                self._font = load_font(weight, style)
            else:
                self._font = self._default_font
        return attrs

    def get_output(self, fname):
        return open(fname, "w")



# This file should be modified by the installer

GLADE="%GLADEFILE%"




if GLADE=="%GLADEFILE%":
    import sys
    import os.path as osp
    # custom configuration for running from the devel directory
    _main_module = sys.modules[__name__]
    _main_dir = osp.dirname( _main_module.__file__)
    _toplevel = osp.abspath(osp.join(_main_dir,".."))
    print "Running from %s" % _toplevel
    GLADE = osp.join(_toplevel, "data", "projedit.glade")



# -*- coding: iso-8859-1 -*-
# Copyright (c) 2004 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Projman - (c)2000-2004 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

__revision__ = "$Id: task.py,v 1.4 2006-02-24 10:39:38 nico Exp $"

from logilab.common.tree import VNode 
from logilab.common.compat import set

from mx.DateTime import now
from projman.lib.constants import *


__all__ = ['MileStone', 'Task']

class TaskNode(VNode):
    """base node for Task and MileStone objects

    Attributes:
      - title
      - date_constraints
      - task_constraints
      - description
    """

    def __init__(self, node_id):
        VNode.__init__(self, node_id)
        self.title = u''
        # list of triplets ('{begin/end}-{at/before/after}-date', mxDate, priority)
        self.date_constraints = set()
        # list of triplets ('{begin/end}-after-{begin/end}', task_id, priority)
        self.task_constraints = set()
        self.description = u''
        self.description_raw = u''
        self.description_format = u'docbook'
        self._progress = None
        self._priority = None
        self.duration = 0
        self.load_type = 0
        self.task_type = None
        self.set_resources = []
        self.can_interrupt = [True, 1] # the integer represent the priority of the constraint
        self.link = None
        self.level = 1
        
    def __repr__(self):
        return "<Task id=%s at %s>" % (self.id, id(self))
    
    def get_task(self, task_id):
        """
        return the task with id task_id
        """
        return self.get_node_by_id(task_id)

    # properties ####################################################
    
    def get_progress(self):
        """used by the 'progress' property"""
        if self.is_leaf():
            if self._progress is None:
                return 0.
            else:
                return self._progress
        else:
            done = [leaf.progress * leaf.maximum_duration()
                    for leaf in self.leaves()]
            return sum(done) / self.maximum_duration()
        
    def set_progress(self, value):
        """used by the 'progress' property"""
        if not self.is_leaf():
            raise AssertionError('Can only set progress on leaves tasks')
        if 0. <= value <= 1.:
            self._progress = value
        else:
            raise ValueError("progress is a percentage (between 0 and 1, got %s)" % value)
    
    progress = property(get_progress, set_progress,
                        doc="task progress (percentage)")

    def get_priority(self):
        """return priority (inherited from parents if None)"""
        if self._priority is None and self.parent:
            return self.parent.priority
        else:
            return self._priority

    def set_priority(self, value):
        self._priority = value

    priority = property(get_priority, set_priority)

    # Editing functions #############################################

    def add_date_constraint(self, constraint_type, date, priority=1):
        """ add a date constraint to the milestone """
        assert constraint_type in DATE_CONSTRAINTS, constraint_type
        self.date_constraints.add((constraint_type, date, priority))

    def add_task_constraint(self, constraint_type, task_id, priority=1):
        """ add a task constraint to the milestone """ 
        assert constraint_type in TASK_CONSTRAINTS, constraint_type
        self.task_constraints.add((constraint_type, task_id, priority))

    # Manipulation functions #############################################
    
    def maximum_duration(self):
        """
        returns the duration of task taking into account all the durations of 
        its children
        /!\ duration is in man-days.
        """
        if self.is_leaf():
            return self.duration
        else:
            assert not self.duration, "Only leaves should have duration ! %s" % self.id 
            return sum([node.maximum_duration() for node in self.children])
    
    def remaining_duration(self):
        """
        return an estimation of the number of days required to close this task
        /!\ duration is in man-days.
        """
        if self.is_leaf():
            if self.progress:
                return self.duration * (1 - self.progress)
            else:
                raise NotImplementedError()
        else:
            return sum([child.remaining_duration() for child in self.children])
        
    def is_ready(self):
        """
        a task is ready when all the tasks it depends on are finished
        """
        for c_type, task_id in self.task_constraints:
            if c_type == BEGIN_AFTER_END:
                task = self.get_task(task_id)
                if task.progress != 1:
                    return False
        return True
        
    def is_finished(self):
        """
        return True if task is finished
        activities is the accomplished activity related to task
        """
        if self.is_leaf():
            return self.progress == 1.
        else:
            for each in self.children:
                if not each.is_finished():
                    return False
            return True
        
    def get_task_constraints(self):
        """
        get all task constraints taking into account the constraints 
        on the parents
        """
        # FIXME: Cache result (or modify in-place)
        if self.parent:
            return self.task_constraints | self.parent.get_task_constraints()
        else:
            return self.task_constraints

    def get_date_constraints(self):
        """
        get the date constraints propagating the appropriate 
        date constraints from parent nodes
        """
        # FIXME: Cache result (or modify in-place)
        constraints = set(self.date_constraints)
        if self.parent:
            for c_type, date, priority in self.parent.get_date_constraints():
                if c_type in (BEGIN_AT_DATE, BEGIN_AFTER_DATE):
                    constraints.add((BEGIN_AFTER_DATE, date, priority))
                elif c_type in (END_AT_DATE, END_BEFORE_DATE):
                    constraints.add((END_BEFORE_DATE, date, priority))
        return constraints

    def check_duration(self):
        """check non valid duration (0)"""
        if self.TYPE != 'milestone' and self.duration == 0:
            raise Exception("non valid task duration for '%s'" %self.id)

    def check_consistency(self):
        """
        check that there are no duplicated task ids
        """
        #check non valid duration (0)
        #if self.TYPE != 'milestone' and self.duration == 0:
        #    raise Exception("non valid task duration for '%s'" %self.id)
        errors = []
        task_ids = set()
        for node in self.flatten():
            if node.id in task_ids:
                errors.append('Duplicate task id: %s' % node.id)
            else:
                task_ids.add(node.id)
        errors += self._check_consistency_date_constraints(self)
        if self.parent :
            errors += self._check_consistency_date_constraints(self.parent)
        return errors

    def _check_consistency_date_constraints(self, other):
        errors = []
        msg = 'Parent task of %s has constraint (%s, %s) which ' \
              'conflicts with its constraints (%s, %s)'
        for other_c_type, other_date, other_priority in other.get_date_constraints():
            if other_c_type in (BEGIN_AT_DATE, BEGIN_AFTER_DATE):
                for c_type, date, priority in self.date_constraints:
                    if date < other_date:
                        errors.append(msg % (self.id, other_c_type,
                                             other_date, c_type, date))
            elif c_type in (END_AT_DATE, END_BEFORE_DATE):
                for c_type, date in self.date_constraints:
                    if date > other_date:
                        errors.append(msg % (self.id, other_c_type,
                                             other_date, c_type, date))
        return errors

class Task(TaskNode):
    """a Task in the Project
    
    A Task instance is identified by its 'id' attribute. It can
    contain other tasks and/or milestone and/or imported projects,
    or it can be a leaf. If it's a leaf, the 'duration' attribute
    indicates the number of days of work needed to perform the
    task.

    Attributes:
      - priority: integer from 0 (lowest) to 9, -1 means inherits
        parent's priority
      - duration: integer estimated duration of the task
      - progress: integer progress of task (in percent)
      - resource_constraints: sequence of tuples (type, id_res),
        can be used without id_res for planning with type of
        resource
      - hide: boolean, used to render partial graphs (e.g. for a
        given resource)
        
    invariant:
        0. <= self.progress <= 1. # percentage of progress
    """ 
    
    TYPE = 'task'

    def __init__(self, t_id):
        TaskNode.__init__(self, t_id)
        self.resource_constraints = set()

    def get_task_type(self):
        if self.task_type:
                return self.task_type
        elif self.parent:
            return self.parent.get_task_type()
        else:
            return set()

    def get_resource_constraints(self):
        """
        get real resource constraints using parent resource constraints
        """
        if not self.task_type:
            if self.resource_constraints:
                return self.resource_constraints
            elif self.parent:
                return self.parent.get_resource_constraints()
            else:
                return set()
        else:
            self.get_task_type()
            return set()
        
    def get_resources(self):
        """
        return a sequence of the resources in resource_constraints
        """
        # according to old resources definition
        self.set_resources = set([id_res for type_c, id_res in
                    self.get_resource_constraints()])
        
        return self.set_resources

    def get_linked_resources(self, set_resources):
        """ return set of resources linked to a task and his children, from all
        resources of set_resources"""
        set_res = set()
        if self.TYPE == 'milestone':
            return set()
        for res in set_resources:
            for leaf in self.leaves():
                if leaf.TYPE == 'milestone':
                    continue
                if leaf.task_type: # according to new definition of resources
                    if leaf.task_type in res.id_role:
                        set_res.add(res.id)
                else: # according to old definition of resources
                    for ctype, res_id in leaf.get_resource_constraints():
                        set_res.add(res_id)
            # if task is a leaf
            if not self.children:
                if self.task_type and not self.children: # according to new definition of resources
                    if self.task_type in res.id_role:
                        set_res.add(res.id)
                else: # according to old definition of resources 
                    for ctype, res_id in self.resource_constraints:
                        set_res.add(res_id)
        return set_res

        
#    def get_resource_dispo(self, res_id):
#        """
#        get the resource disponibility required for this task
#        """
#        for type_c, id_res in self.get_resource_constraints():
#            if id_res == res_id:    
#                return usage
#        return 0

    def add_resource_constraint(self, resource_type, resource_id):
        # FIXME what about node <constraint-resource idref="toto"/>
        self.resource_constraints.add((resource_type, resource_id))


class MileStone(TaskNode):
    """A milestone in the project
    
    A milestone represents an event occuring at a specific date, but
    without any specific information about duration or timing.

    Subclasses of MileStone are Task and Project.

    """
    
    TYPE = 'milestone'

    def append(self, node):
        raise AssertionError("a milestone should always be a leaf")

    def get_progress(self):
        if self.is_ready():
            return 1.
        return 0.
    
    progress = property(get_progress, doc="milestone progress (percentage)")

    def get_resource_constraints(self):
        return set()



# Copyright (c) 2004-2007 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
projman is a python package to schedule and transform xml project file to
gantt diagram, resource diagrams, etc.
"""

__revision__ = "$Id: __init__.py,v 1.12 2005-11-10 11:34:18 arthur Exp $"

import math

ENCODING = "ISO-8859-1"
LOG_CONF = "logging.conf"

DAY_WEEK = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']

# In order to format monetary numbers, we could think of standard
# package locale, but it does not include spaces when using fr_FR.
#
#  >>> import locale
#  >>> locale.setlocale(locale.LC_ALL, '')
#  fr_FR
#  >>> locale.format("%.2f", 121212.01, grouping=True)
#  121212,01
def format_monetary(str_number, nb_decimal=2, space_gap=3):
    """takes a string or number and return it formatted in a monetary way. 
    ex: 121212.01 as u'121 212,01'

    nb_decimal sets the number of decimal to display space_gap sets
    
    the gap (ie, number of numbers) between two of the spaces. It is 3
    in the previous example.
    
    raises a TypeError in case of bad format
    """
    # FIXME REFACTOR to
    # fmt = reversed(str(intval))
    # final = []
    # while fmt:
    #     final.append(fmt[:space_gap])
    #     fmt = fmt[space_gap:]
    # output = '\xA0'.join(reversed(final))
    # 
    # FIXME: decide whether or not we should call round here
    fractval, intval = math.modf(float(str_number))
    intval = int(intval)
    if nb_decimal > 0:
        fract_part = round(fractval * 100)
        if fract_part == 100:
            intval += 1
            fract_part = 0
    parts = []
    div = 10**space_gap
    if space_gap: # space_gap = 0 would mean div=1, and infinite loop
        # Build the int part's chunks
        while intval >= div:
            intval, mod = divmod(intval, div)
            parts.insert(0, '%%0%dd' % space_gap % mod)
    parts.insert(0, str(intval))
    int_part = u'\xA0'.join(parts)
    if nb_decimal:
        fmt = '%%s,%%0%dd' % nb_decimal
        return fmt % (int_part, fract_part)
    return int_part

   

def extract_extension(name):
    """returns a couple (file_name, ext) from file_name.ext"""
    # XXX is this different from os.path.basename ??
    splitted = name.split('.', 1)
    if len(splitted) == 2:
        return tuple(splitted)
    else:
        return (name, '')



# -*- coding: utf-8 -*-
# Copyright (c) 2000-2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
reader generate a model from xml file (see dtd/project.dtd)
"""
from curses.ascii import isascii
from projman.readers.base_reader import AbstractXMLReader
from projman.readers.projman_checkers import ProjectChecker, ScheduleChecker, ResourcesChecker
from projman.readers.projman_checkers import TasksChecker
from projman.readers.projman_checkers import iso_date, iso_time
from os.path import dirname, abspath, isabs, join
from logilab.common.table import Table
from docutils.core import publish_string
from logilab.doctools.rest_docbook import FragmentWriter
from logilab.common.textutils import colorize_ansi
from projman.lib._exceptions import ProjectValidationError, MalformedProjectFile, MalformedId
from projman import DAY_WEEK
from projman.lib.constants import LOAD_TYPE_MAP, TASK_MILESTONE
docbook_writer = FragmentWriter()

try:
    import xml.etree.ElementTree as ET
except ImportError:
    import elementtree.ElementTree as ET

def js(txt):
    """join and split"""
    return u' '.join(txt.split())


class ProjectXMLReader(AbstractXMLReader) :

    def __init__(self, src, task_root=None, skip_schedule=False):
        AbstractXMLReader.__init__(self)
        self.source = src
        self.task_root = task_root
        self.skip_schedule = skip_schedule
        self.project = self._factory.create_project()
        self.files = {'tasks': None,
                      'schedule': None,
                      'resources': None,
                      'activities': None,
                      }

    def read(self):
        if isinstance(self.source, str):
            tree = ET.parse( file(self.source) )
            filename = self.source
            base_uri = dirname(abspath(filename))
        elif isinstance(self.source, file):
            tree = ET.parse(self.source)
            filename = 'input_stream'
            base_uri = ''
        elif isinstance(self.source, ET.ElementTree):
            tree = self.source
            filename = 'input_tree'
            base_uri = ''
        else:
            raise ValueError('unknown source %s' % self.source)
        return self.fromTree(tree, filename, base_uri), self.files
        
    def get_file(self, tree, ftype, default=None):
        node = tree.find("import-"+ftype)
        if node is None:
            return default
        fname = node.get("file", default)
        self.files[ftype] = fname
        if not isabs(fname):
            fname = join(self._base_uris[-1], fname)
        return fname

    def fromTree(self, tree, filename="input_stream", base_uri=''):
        self._base_uris.append(base_uri)
        checker = ProjectChecker()
        if not checker.validate(tree, filename):
            raise MalformedProjectFile(str(checker))
        sched = self.get_file(tree, "schedule")
        tasks = self.get_file(tree, "tasks")
        rsrc = self.get_file(tree, "resources")
        act = self.get_file(tree, "activities")
        self.project.root_task = self.read_tasks(tasks)
        self.project.resource_set = self.read_resources(rsrc)
        self.project.resource_role_set = self.read_resource_role(rsrc)
        self.project.add_activities( self.read_activities(act) )
        if sched and not self.skip_schedule:
            try:
                file(sched,"r")
            except IOError:
                print colorize_ansi("WARNING!", "red"), \
                      " schedule file '%s' declared in project but file is missing. " \
                      "Command completed without scheduling information."% filename
            else:
                self.read_schedule( sched )
        return self.project

    def read_schedule(self, fname):
        schedule = ET.parse( fname )
        checker = ScheduleChecker()
        if not checker.validate(schedule, fname):
            raise MalformedProjectFile(str(checker))
        
        activities = Table(default_value=None,
                           col_names=['begin', 'end', 'resource', 'task',
                                      'usage', 'src'])
        tasks = Table(default_value=None,
                      col_names=['begin', 'end', 'status', 'cost', 'unit'])
        costs = Table(default_value=None,
                      col_names=['task', 'resource', 'cost', 'unit'])

        for task in schedule.findall("task"):
            t_id = task.get('id')
            tasks.create_row( t_id )
            global_cost = task.find("global-cost")
            tasks.set_cell_by_ids(t_id, 'unit', global_cost.get('unit'))
            tasks.set_cell_by_ids(t_id, 'cost', float(global_cost.text) )
            tasks.set_cell_by_ids(t_id, 'status', task.find('status').text )
            for cd in task.findall("contraint-date"):
                date = iso_date( cd.text )
                if cd.get('type') == 'begin-at-date':
                    tasks.set_cell_by_ids( t_id, 'begin', date )
                elif cd.get('type') == 'end-at-date':
                    tasks.set_cell_by_ids( t_id, 'end', date )
            for ct in task.findall("constraint-task"):
                pass
                #raise NotImplementedError("constraint-task not implemented in schedule")
            for report in task.findall("report-list/report"):
                activities.append_row( (iso_date(report.get('from')),
                                             iso_date(report.get('to')),
                                             report.get('idref'),
                                             t_id,
                                             float(report.get('usage')) ) )
            for cost in task.findall("costs_list/cost"):
                costs.append_row( (t_id, cost.attrib['idref'],float(cost.text),None) )

        for milestone in schedule.findall("milestone"):
            t_id = milestone.get('id')
            cd = milestone.find("constraint-date")
            date = iso_date( cd.text )
            self.project.milestones[t_id] = date

        self.project.add_schedule(activities)
        self.project.tasks = tasks
        self.project.costs = costs

    def id_checker(self, id_):
        checker = True
        for c in id_:
            checker = checker and  isascii(c)
        return checker

    def read_tasks(self, fname):
        tasks = ET.parse( fname )
        checker = TasksChecker()
        if not checker.validate(tasks, fname):
            raise MalformedProjectFile(str(checker))
        self.tasks_file = fname
        rt = tasks.getroot()
        if not self.id_checker(rt.get("id")):
            raise MalformedId("Task %s has non-ascii ID." %(t.get("id")))
        if self.task_root and rt.get("id") != self.task_root:
            for t in rt.findall(".//task"):
                if not self.id_checker(t.get("id")):
                    raise MalformedId("Task %s has non-ascii ID." %(t.get("id")))
                if t.get("id") == self.task_root:
                    rt = t
                    break
            else:
                raise RuntimeError("Task root %s not found" % self.task_root)
        return self.read_task(rt, 0)

    def read_task(self, task, niveau):
        t = self._factory.create_task( task.get("id") )
        t.level = niveau
        if not self.id_checker(task.get("id")):
            raise MalformedId("Task %s has non-ascii ID." %(t.get("id")))
        if task.get("resource-role"):
            t.task_type = task.get("resource-role")
        self.task_milestone_common( t, task )
        for child in task:
            if child.tag == 'constraint-interruptible':
                if child.get('type') == 'False':
                    t.can_interrupt[0] = False
                if child.get('priority'):
                    t.can_interrupt[1] = int(child.get('priority'))
            elif child.tag == "progress":
                t.progress = float(child.text)
            elif child.tag == "priority":
                t.priority = int(child.text)
            elif child.tag == "task":
                t.append( self.read_task( child, niveau+1))
            elif child.tag == "milestone":
                t.append( self.read_milestone(child) )
        return t

    def read_milestone(self, mstone):
        if not self.id_checker(mstone.get("id")):
            raise MalformedProjectFile()
        m = self._factory.create_milestone( mstone.get("id") )
        self.task_milestone_common( m, mstone )
        return m

    def task_milestone_common(self, t, task):
        t.title = js(unicode(task.find("label").text))
        load_type = LOAD_TYPE_MAP[ task.get("load-type", "shared").lower() ]
        if task.tag=="milestone":
            load_type = TASK_MILESTONE
        t.load_type = load_type
        t.duration = float(task.get("load", "0"))
        for link in task.findall('link'):
            t.link = link.get('url')
        for cd in task.findall("constraint-date"):
            if cd.get("priority"):
                priority = cd.get("priority")
            else:
                priority = 1
            t.add_date_constraint( cd.get("type"), iso_date( cd.text ), priority )
        for ct in task.findall("constraint-task"):
            if ct.get("priority"):
                priority = ct.get("priority")
            else:
                priority = 1
            t.add_task_constraint( ct.get("type"), ct.get("idref"), priority)
        for cr in task.findall("constraint-resource"):
            t.add_resource_constraint( cr.get("type"), cr.get("idref"))
        
        desc = task.find("description")
        txt_fmt = "none"
        if desc is None:
            txt = u""
            raw_txt = u""
        else:
            if desc.get("format")=="docbook":
                txt_fmt = "docbook"
            txt = unicode(desc.text) or u""
            for n in desc:
                txt+=unicode(ET.tostring(n,"utf-8"),"utf-8")
            raw_txt = txt
            if desc.get("format")=="rest":
                txt_fmt = "rest"
                txt = publish_string(txt,
                                     settings_overrides={'output_encoding': 'unicode'},
                                     writer=docbook_writer,
                                     source_path=self.tasks_file + "<%s>"%t.id)
        t.description = txt
        t.description_raw = raw_txt
        t.description_format = txt_fmt


    def read_resource_definition(self, res_node):
        if not self.id_checker(res_node.get("id")):
            raise MalformedId("Task %s has non-ascii ID." %(t.get("id")))
        res = self._factory.create_resource( res_node.get('id'), u'',
                                             res_node.get('type'), u'' )
        for n in res_node:
            if n.tag == 'label':
                res.name = unicode(n.text)
            elif n.tag == 'hourly-rate':
                res.hourly_rate[0] = float(n.text)
                res.hourly_rate[1] = n.get('unit','euros')
            elif n.tag == 'use-calendar':
                res.calendar = n.get('idref')
            elif n.tag == 'role':
                res.id_role.append(n.get('idref'))
        return res

    def read_resource_role(self, fname):
        tree = ET.parse(fname)
        root_node = tree.getroot()
        res_role_set = self._factory.create_resource_role_set('all_resource_role')
        for res_role in root_node.findall('resource-role'):
            if not self.id_checker(res_role.get("id")):
                raise MalformedId("Task %s has non-ascii ID." %(t.get("id")))
            res = self._factory.create_resource_role( res_role.get('id'), u'')
            res.hourly_cost = float(res_role.get("hourly-cost"))
            res.unit = res_role.get("cost-unit")
            for n in res_role._children:
                if n.tag == 'label':
                    res.name = unicode(n.text)
            res_role_set.append(res)
        return res_role_set

    def read_calendar_definition(self, cal_node, parent_cal=None):
        cal = self._factory.create_calendar( cal_node.get('id') )
        cal.type_working_days = {}
        cal.type_nonworking_days = {}
        for n in cal_node:
            if n.tag == "label":
                cal.name = unicode(n.text)
            elif n.tag == "day-types":
                cal.default_day_type = n.get('default')
                for day_type in n.findall('day-type'):
                    day_id = day_type.get('id')
                    day_type_label = unicode(day_type.find('label').text)
                    intervals = []
                    for interval in day_type.findall('interval'):
                        from_time = iso_time( interval.get('start') )
                        to_time = iso_time( interval.get('end') )
                        intervals.append( (from_time, to_time) )
                    cal.day_types[day_id] = [day_type_label, intervals]
            elif n.tag == "day":
                day_type = n.get('type')
                data = n.text
                if data in DAY_WEEK:
                    cal.weekday[data] = day_type
                elif len(data) < 8:
                    cal.national_days.append(tuple([int(item) for item in data.split('-')]))
                else:
                    date = iso_date(data)
                    cal.add_timeperiod(date, date, day_type)
            elif n.tag == "timeperiod":
                from_date = iso_date( n.get('from') )
                to_date = iso_date( n.get('to') )
                type_name = n.get('type')
                cal.add_timeperiod(from_date, to_date, type_name)
            elif n.tag == 'start-on':
                cal.start_on = iso_date(n.text)
            elif n.tag == 'stop-on':
                cal.stop_on = iso_date(n.text)
            elif n.tag == 'calendar':
                subcal = self.read_calendar_definition(n, parent_cal=cal)
                cal.append(subcal)
        return cal

    def read_resources(self, fname):

        tree = ET.parse(fname)
        checker = ResourcesChecker()
        if not checker.validate(tree, fname):
            raise MalformedProjectFile(str(checker))
        root_node = tree.getroot()
        res_set = self._factory.create_resourcesset('all_resources')
        for res_node in root_node.findall('resource'):
            res = self.read_resource_definition( res_node )
            res_set.append(res)
        for cal_node in root_node.findall('calendar'):
            cal = self.read_calendar_definition( cal_node )
            res_set.append(cal)
        return res_set
    
    def read_activities(self, fname):
        tree = ET.parse(fname)
        root_node = tree.getroot()
        activities = []
        for reports in root_node.findall('reports-list'):
            task_id = reports.get('task-id')
            for report in reports.findall('report'):
                res_id = report.get('idref')
                begin = iso_date( report.get('from') )
                end = iso_date( report.get('to') )
                usage = float( report.get('usage') )
                activities.append( (begin, end, res_id, task_id, usage) )
        return activities



# Copyright (c) 2004 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
projman is a python package to schedule and transform xml project file to
gantt diagram, resource diagrams, etc.
"""

import os.path as osp
from projman import extract_extension

DATADIR = osp.abspath(osp.join(osp.dirname(__file__),'data'))

PYGANTT_PROJECT = osp.join(DATADIR, "pygantt_planif.xml")

TAR_TARED_PROJMAN = "projman.prj"
TAR_TARED_SCHEDULED_PROJMAN = "scheduled_projman.prj"
TAR_PROJMAN = "projman.xml"
TAR_TASK_FILE = "tasks.xml"
TAR_RESOURCE_FILE = "resources.xml"
TAR_ACTIVITY_FILE = "activities.xml"

XML_TARED_PROJMAN = "trivial_projman.prj"
XML_PROJMAN ="trivial_projman.xml"
XML_SCHEDULED_PROJMAN ="trivial_scheduled_projman.xml"
XML_SCHEDULED_PROJMAN_FULL ="full_scheduled_projman.xml"
XML_TASK_FILE = "three_tasked_project.xml"
XML_RESOURCE_FILE ="three_resourced_list.xml"
XML_ACTIVITY_FILE = "three_activities.xml"



# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
external access point to projman base classes
"""

__revision__ = "$Id: __init__.py,v 1.3 2005-09-06 17:07:21 nico Exp $"

from abstract import AbstractRenderer, AbstractDrawer
from gantt import GanttRenderer
from resource import ResourcesRenderer
from ganttresource import GanttResourcesRenderer
from pil import PILHandler, FORMATS as PIL_FORMATS
from svg import SVGHandler

def HandlerFactory(format):
    format = format.upper()
    if format in PIL_FORMATS:
        return PILHandler(format)
    if format == 'SVG':
        return SVGHandler(format)
    raise NotImplementedError("Format '%s' not supported" % format )



# -*- coding: iso-8859-1 -*-
"""
unit tests for module projman.lib.Milestone

Projman - (c)2000-2002 LOGILAB <contact@logilab.fr> - All rights reserved.
Home: http://www.logilab.org/projman

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

from mx.DateTime import DateTime
from logilab.common.testlib import TestCase, unittest_main

from projman.lib import MileStone
from projman.lib.constants import BEGIN_AFTER_END, BEGIN_AFTER_BEGIN, \
     END_AFTER_END, END_AFTER_BEGIN, \
     BEGIN_AFTER_DATE, BEGIN_AT_DATE, BEGIN_BEFORE_DATE, \
     END_AFTER_DATE, END_AT_DATE, END_BEFORE_DATE

class MilestoneTest(TestCase):
    """Test all low-level function on Milestones.
    """
    def setUp(self):
        """ called before each test from this class """
        self.stone = MileStone('Test')
        self.child_stone = MileStone('Test_child')
        self.unicode_stone = MileStone(u'accentuée')
        self.unicode_stone.description = u"autre chaîne unicode non-ASCII"
        # set of dates
        self.date_today = DateTime(2004, 10, 7)
        self.date_tomorrow = DateTime(2004, 10, 8)
        self.date_yesterday = DateTime(2004, 10, 6)
        # set of date constraints
        self.at_constraints =  {BEGIN_AT_DATE: self.date_yesterday,
                                END_AT_DATE: self.date_tomorrow}
        self.before_constraints =  {BEGIN_BEFORE_DATE: self.date_today,
                                    END_BEFORE_DATE: self.date_tomorrow}
        self.after_constraints =  {BEGIN_AFTER_DATE: self.date_yesterday,
                                   END_AFTER_DATE: self.date_today}
        self.before_after_constraints =  {BEGIN_BEFORE_DATE: self.date_today,
                                          END_AFTER_DATE: self.date_yesterday}
        # set of id
        self.good_id = 'good_id'
        self.other_id = 'other_id'
        self.another_id = 'another_id'
        self.last_id = 'last_id'
        
        
    def test_add_date_constraint(self):
        """No validation over which parameter is given.
        """
        self.stone.add_date_constraint(BEGIN_AT_DATE, self.date_today)
        self.assertRaises(AssertionError,
                          self.stone.add_date_constraint, 'fake', self.date_tomorrow)
        self.assertSetEqual(self.stone.get_date_constraints(),
                            [(BEGIN_AT_DATE, self.date_today, 1)])
        
    def test_add_task_constraint(self):
        """No validation over which parameter is given.
        """
        self.stone.add_task_constraint(BEGIN_AFTER_BEGIN, self.good_id)
        self.assertRaises(AssertionError, self.stone.add_task_constraint, 'fake', 'dumb_id')
        self.assertSetEqual(self.stone.get_task_constraints(),
                          [(BEGIN_AFTER_BEGIN,self.good_id, 1)])

#     def __test_get_range_at(self):
#         """get date range after setting constraints of type 'at' with 
#         """
#         for cst, date in self.at_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         self.assertEquals(self.stone.get_date_range(),
#                           (self.date_yesterday, self.date_tomorrow))
        
#     def __test_get_range_before(self):
#         """get date range after setting constraints of type 'before' with today, tomorrow
#         """
#         for cst, date in self.before_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         self.assertEquals(self.stone.get_date_range(),
#                           (self.date_today, self.date_tomorrow))
        
#     def __test_get_range_after(self):
#         """get date range after setting constraints of type 'after' with yesterday, today
#         """
#         for cst, date in self.after_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         self.assertEquals(self.stone.get_date_range(),
#                           (self.date_yesterday, self.date_today))
        
#     def __test_get_range_before_after(self):
#         """get date range after setting constraints of type 'before, after' with today, yesterday
#         """
#         for cst, date in self.before_after_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         self.assertEquals(self.stone.get_date_range(),
#                           (self.date_today, self.date_yesterday))

#     def ___test_get_task_constraints(self):
#         """Create a child and checks inheritance of task constraints
#         """
#         # set up Milestone
#         constraints = {BEGIN_AFTER_BEGIN: [self.good_id, self.other_id],
#                        END_AFTER_END: [self.another_id]}
#         for key, values in constraints.items():
#             for value in values:
#                 self.stone.add_task_constraint(key, value)
#         # child
#         child_constraints =  {BEGIN_AFTER_END: [self.another_id],
#                               END_AFTER_END : [self.last_id]}
#         for key, values in child_constraints.items():
#             for value in values:
#                 self.child_stone.add_task_constraint(key, value)
#         # create parent
#         self.stone.append(self.child_stone)
#         # test getter
#         all_constraints = {BEGIN_AFTER_BEGIN: [self.good_id, self.other_id],
#                            END_AFTER_END: [self.another_id, self.last_id],
#                            BEGIN_AFTER_END: [self.another_id]}
#         result = self.child_stone.get_task_constraints()
#         for key in all_constraints.keys():
#             all_constraints[key].sort()
#             result[key].sort()
#             self.assertEquals(all_constraints[key], result[key])
        
#     def ___test_get_date_constraints(self):
#         """Create a child and checks inheritance of date constraints
#            BEWARE: begin-at-date & end-at-date ar inherited as
#            begin-after-date & end-before-date
#         """
#         # milestone
#         for cst, date in self.at_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         for cst, date in self.before_constraints.items():
#             self.stone.add_date_constraint(cst, date)
#         for cst, date in self.before_after_constraints.items():
#             self.child_stone.add_date_constraint(cst, date)
#         # create parent
#         self.stone.append(self.child_stone)
#         # test getter
#         all_constraints = {BEGIN_BEFORE_DATE: self.date_today,
#                            END_BEFORE_DATE: self.date_tomorrow,
#                            BEGIN_BEFORE_DATE: self.date_today,
#                            END_AFTER_DATE: self.date_yesterday}
#         result = self.child_stone.get_date_constraints()
#         for key in all_constraints.keys():
#             self.assertEquals(all_constraints[key], result[key])
#         for key in [BEGIN_AT_DATE, END_AT_DATE]:
#             self.assertRaises(KeyError, result.__getitem__, key)

if __name__ == '__main__':
    unittest_main()
